""" BlueSky Datalogger """

# ToDo: Add description in comments

import numbers
import itertools
from datetime import datetime
import numpy as np
from bluesky import settings, stack
from bluesky.core import varexplorer as ve
import bluesky as bs
from bluesky.stack import command

# Register settings defaults
settings.set_variable_defaults(log_path='output')

logprecision = '%.8f'

# Dict to contain the definitions of periodic loggers
periodicloggers = dict()

# Dict to contain all loggers (also the periodic loggers)
allloggers = dict()


@command(name='CRELOG')
def crelogstack(name: 'txt', dt: float = None, header: 'string' = ''):
    """ Create a new data logger.

        Arguments:
        - name: The name of the logger
        - dt: The logging time interval. When a value is given for dt
              this becomes a periodic logger.
        - header: A header text to put at the top of each log file
    """
    if name in allloggers:
        return False, f'Logger {name} already exists'

    crelog(name, dt, header)
    return True, f'Created {"periodic" if dt else ""} logger {name}'


def crelog(name, dt=None, header=''):
    """ Create a new logger. """
    allloggers[name] = allloggers.get(name, CSVLogger(name, dt or 0.0, header))
    if dt:
        periodicloggers[name] = allloggers[name]

    return allloggers[name]


def update():
    """ This function writes to files of all periodic logs by calling the appropriate
    functions for each type of periodic log, at the approriate update time. """
    for log in periodicloggers.values():
        log.log()


def reset():
    """ This function closes all logs. It is called when simulation is
    reset and at quit. """

    CSVLogger.simt = 0.0

    # Close all logs and remove reference to its file object
    for log in allloggers.values():
        log.reset()


def makeLogfileName(logname, prefix: str = ''):
    timestamp = datetime.now().strftime('%Y%m%d_%H-%M-%S')
    if prefix == '' or prefix.lower() == stack.get_scenname().lower():
        fname = "%s_%s_%s.log" % (logname, stack.get_scenname(), timestamp)
    else:
        fname = "%s_%s_%s_%s.log" % (logname, stack.get_scenname(), prefix, timestamp)
    return bs.resource(settings.log_path) / fname


def col2txt(col, nrows):
    if isinstance(col, (list, np.ndarray)):
        if isinstance(col[0], numbers.Integral):
            ret = np.char.mod('%d', col)
        elif isinstance(col[0], numbers.Number):
            ret = np.char.mod(logprecision, col)
        else:
            ret = np.char.mod('%s', col)
        if len(ret.shape) > 1:
            for el in ret.T:
                yield el
        else:
            yield ret
    elif isinstance(col, numbers.Integral):
        yield nrows * ['%d' % col]
    elif isinstance(col, numbers.Number):
        yield nrows * [logprecision % col]
    # The input is not a number
    else:
        yield nrows * [col]


class CSVLogger:
    def __init__(self, name, dt, header):
        self.name = name
        self.file = None
        self.fname = None
        self.dataparents = []
        self.header = header.split('\n')
        self.tlog = 0.0
        self.selvars = []

        # In case this is a periodic logger: log timestep
        self.dt = dt
        self.default_dt = dt

        # Register a command for this logger in the stack
        stackcmd = {name: [
            name + ' ON/OFF,[dt] or ADD [FROM parent] var1,...,varn',
            '[txt,float/word,...]', self.stackio, name + " data logging on"]
        }
        stack.append_commands(stackcmd)

    def write(self, line):
        self.file.write(bytearray(line, 'ascii'))

    def setheader(self, header):
        self.header = header.split('\n')

    def setdt(self, dt):
        self.dt = dt
        self.default_dt = dt

    def addvars(self, selection):
        selvars = []
        while selection:
            parent = ''
            if selection[0].upper() == 'FROM':
                parent = selection[1] + '.'
                del selection[0:2]
            variables = list(itertools.takewhile(
                lambda i: i.upper() != 'FROM', selection))
            selection = selection[len(variables):]
            for v in variables:
                varobj = ve.findvar(parent + v)
                if varobj:
                    selvars.append(varobj)
                else:
                    return False, f'Variable {v} not found'

        self.selvars = selvars
        return True

    def open(self, fname):
        if self.file:
            self.file.close()
        self.file = open(fname, 'wb')
        # Write the header
        for line in self.header:
            self.file.write(bytearray('# ' + line + '\n', 'ascii'))
        # Write the column contents
        columns = ['simt']
        for v in self.selvars:
            columns.append(v.varname)
        self.file.write(
            bytearray('# ' + str.join(', ', columns) + '\n', 'ascii'))

    def isopen(self):
        return self.file is not None

    def log(self, *additional_vars):
        if self.file and bs.sim.simt >= self.tlog:
            # Set the next log timestep
            self.tlog += self.dt

            # Make the variable reference list
            varlist = [bs.sim.simt]
            varlist += [v.get() for v in self.selvars]
            varlist += additional_vars

            # Get the number of rows from the first array/list
            nrows = 1
            for v in varlist:
                if isinstance(v, (list, np.ndarray)):
                    nrows = len(v)
                    break
            if nrows == 0:
                return
            # Convert (numeric) arrays to text, leave text arrays untouched
            txtdata = [
                txtcol for col in varlist for txtcol in col2txt(col, nrows)]

            # log the data to file
            np.savetxt(self.file, np.vstack(txtdata).T,
                       delimiter=',', newline='\n', fmt='%s')

    def start(self, prefix: str = ''):
        """ Start this logger. """
        self.tlog = bs.sim.simt
        self.fname = makeLogfileName(self.name, prefix)
        self.open(self.fname)

    def reset(self):
        self.dt = self.default_dt
        self.tlog = 0.0
        self.fname = None
        if self.file:
            self.file.close()
            self.file = None

    def listallvarnames(self):
        return str.join(', ', (v.varname for v in self.selvars))

    def stackio(self, *args):
        if len(args) == 0:
            text = 'This is '
            if self.name in periodicloggers:
                text += 'a periodic logger, with an update interval of %.2f seconds.\n' % self.dt
            else:
                text += 'a non-periodic logger.\n'

            text += 'with variables: ' + self.listallvarnames() + '\n'
            text += self.name + ' is ' + ('ON' if self.isopen() else 'OFF') + \
                '\nUsage: ' + self.name + \
                ' ON/OFF,[dt] or ADD [FROM parent] var1,...,varn'
            return True, text
            # TODO: add list of logging vars
        elif args[0] == 'ON':
            if len(args) > 1:
                if isinstance(args[1], float):
                    self.dt = args[1]
                else:
                    return False, 'Turn ' + self.name + ' on with optional dt'
            self.start()

        elif args[0] == 'OFF':
            self.reset()

        elif args[0] == 'ADD':
            return self.addvars(list(args[1:]))

        return True



'''BlueSky global configuration module'''
import re
import inspect
from bluesky.pathfinder import resource


# Default config file path
_cfgfile = 'settings.cfg'


def __getattr__(name):
    raise AttributeError(f'Variable \'{name}\' not found in settings. '
                         'You should specify a default value with '
                         'bluesky.settings.set_variable_defaults()'
                         'to avoid this error.')


def init(cfgfile=''):
    '''Initialize configuration.
       Import config settings from settings.cfg if this exists, if it doesn't
       create an initial config file'''

    # If no alternative is given we'll use the default config file
    if not cfgfile:
        cfgfile = _cfgfile

    # Store name of config file
    globals()['_cfgfile'] = resource(cfgfile)

    # Read the configuration file
    print(f'Reading config from {_cfgfile}')
    with open(_cfgfile) as fin:
        config = fin.read()
    # BlueSky resources have been moved since v2022.9.19. Update config file if necessary
    matches = re.findall(r'.+_path.+=.*data[/\\].*', config)
    if matches:
        print(f'Old-style resource paths found in {_cfgfile}. Do you want to update it')
        print('with the following changes?')
        for line in matches:
            var, p = line.split('=')
            np = p.strip().replace('data/', '').replace('data\\', '')
            print(f'{var.strip()}: from {p.strip()} to {np}')
        resp = input('Make these changes? [Y/n]: ')
        if (not resp) or 'y' in resp.lower():
            # Update settings and store to file
            print(f'Updating {_cfgfile}')
            config = config.replace('data/', '').replace('data\\', '')
            with open(_cfgfile, 'w') as fout:
                fout.write(config)
        else:
            print('If bluesky doesn\'t load correctly, please update your config file manually')

    exec(compile(config.replace('\\', '/'), _cfgfile, 'exec'), globals())

    return True


_settings_hierarchy = dict()
_settings = list()
def set_variable_defaults(**kwargs):
    ''' Register a default value for a configuration variable. Use this functionality
        in plugins to make sure that configuration variables are available upon usage.

        Example:
            from bluesky import settings
            settings.set_variable_defaults(var1=1.0, var2=[1, 2, 3])

            This will make settings.var1 and settings.var2 available, with the
            provided default values.'''
    for key, value in kwargs.items():
        if key not in globals():
            globals()[key] = value
        else:
            kwargs[key] = globals()[key]
        if key not in _settings:
            _settings.append(key)

    # Keep track of who this variable belongs to
    callertree = inspect.currentframe().f_back.f_globals['__name__'].split('.')
    tree = _settings_hierarchy
    visited = set()
    for loc in callertree:
        if loc in visited:
            continue
        if loc not in tree:
            tree[loc] = dict()
        tree = tree[loc]
        visited.add(loc)
    tree.update(kwargs)

def save(fname=None, changes=None):
    ''' Save BlueSky configuration to file. '''
    # Apply any changes that are passed for saving
    if changes:
        globals().update(changes)
    # Make a copy of the list of settings
    allsettings = list(_settings)
    # Write to a specified file if passed, else current config file
    fname = resource(fname or _cfgfile)
    # Get config file formatting from file to be updated if it exists, else get
    # it from the config file template
    configsrc = fname if fname.is_file() else resource('default.cfg')
    with open(configsrc) as f:
        lines = f.readlines()

    with open(fname, 'w') as file_out:
        # first write all lines following the format of the source file
        for line in lines:
            key = (re.findall(r'^(\w+)\s*=?.*', line.strip()) + [''])[0]
            if key in allsettings:
                allsettings.pop(allsettings.index(key))
                value = globals()[key]
                if isinstance(value, str):
                    file_out.write(f'{key} = \'{value}\'\n')
                else:
                    file_out.write(f'{key} = {value}\n')
            else:
                file_out.write(line)
        # Then write any remaining additional settings
        file_out.write('\n')
        for key in allsettings:
            value = globals()[key]
            if isinstance(value, str):
                file_out.write(f'{key} = \'{value}\'\n')
            else:
                file_out.write(f'{key} = {value}\n')

    return True, f'Saved settings to {fname}'



"""Area filter module"""
from weakref import WeakValueDictionary
import numpy as np
from matplotlib.path import Path
try:
    from rtree.index import Index
except (ImportError, OSError):
    print('Warning: RTree could not be loaded. areafilter get_intersecting and get_knearest won\'t work')
    class Index:
        ''' Dummy index class for installations where rtree is missing
            or doesn't work.
        '''
        @staticmethod
        def intersection(*args, **kwargs):
            return []

        @staticmethod
        def nearest(*args, **kwargs):
            return []

        @staticmethod
        def insert(*args, **kwargs):
            return

        @staticmethod
        def delete(*args, **kwargs):
            return


from bluesky.stack import commandgroup
from bluesky.tools.geo import kwikdist
from bluesky.network.publisher import StatePublisher


# Dictionary of all basic shapes (The shape classes defined in this file) by name
basic_shapes = dict()
# Publisher object to manage publishing of states to clients
polypub = StatePublisher('POLY', collect=True)


@polypub.payload
def puball():
    return dict(polys={name: poly.raw for name, poly in basic_shapes.items()})


@commandgroup(annotations='txt,color', aliases=('COLOR', 'COL'))
def colour(name, r, g, b):
    ''' Set custom color for visual objects. '''
    poly = basic_shapes.get(name)
    if poly:
        poly.color = (r, g, b)
        polypub.send_update(polys={name:dict(color=poly.color)})
        return True
    return False, 'No shape found with name ' + name


def hasArea(areaname):
    """Check if area with name 'areaname' exists."""
    return areaname in basic_shapes


def getArea(areaname):
    ''' Return the area object corresponding to name '''
    return basic_shapes.get(areaname, None)


def defineArea(name, shape, coordinates, top=1e9, bottom=-1e9):
    """Define a new area"""
    if name == 'LIST':
        if not basic_shapes:
            return True, 'No shapes are currently defined.'
        else:
            return True, 'Currently defined shapes:\n' + \
                ', '.join(basic_shapes)
    if not coordinates:
        if name in basic_shapes:
            return True, str(basic_shapes[name])
        else:
            return False, f'Unknown shape: {name}'
    if shape == 'BOX':
        basic_shapes[name] = Box(name, coordinates, top, bottom)
    elif shape == 'CIRCLE':
        basic_shapes[name] = Circle(name, coordinates, top, bottom)
    elif shape[:4] == 'POLY':
        basic_shapes[name] = Poly(name, coordinates, top, bottom)
    elif shape == 'LINE':
        basic_shapes[name] = Line(name, coordinates)

    # Pass the shape on to the connected clients
    polypub.send_update(polys={name:dict(shape=shape, coordinates=coordinates)})

    return True  #, f'Created {shape} {name}'


def checkInside(areaname, lat, lon, alt):
    """ Check if points with coordinates lat, lon, alt are inside area with name 'areaname'.
        Returns an array of booleans. True ==  Inside"""
    if areaname not in basic_shapes:
        return np.zeros(len(lat), dtype=bool)
    area = basic_shapes[areaname]
    return area.checkInside(lat, lon, alt)

def deleteArea(areaname):
    """ Delete area with name 'areaname'. """
    if areaname in basic_shapes:
        basic_shapes.pop(areaname)
        # bs.scr.objappend('', areaname, None)
        polypub.send_delete(polys=[areaname])

def reset():
    """ Clear all data. """
    basic_shapes.clear()
    Shape.reset()


def get_intersecting(lat0, lon0, lat1, lon1):
    ''' Return all shapes that intersect with a specified rectangular area.

        Arguments:
        - lat0/1, lon0/1: Coordinates of the top-left and bottom-right corner
          of the intersection area.
    '''
    items = Shape.areatree.intersection((lat0, lon0, lat1, lon1))
    return [Shape.areas_by_id[i.id] for i in items]


def get_knearest(lat0, lon0, lat1, lon1, k=1):
    ''' Return the k nearest shapes to a specified rectangular area.

        Arguments:
        - lat0/1, lon0/1: Coordinates of the top-left and bottom-right corner
          of the relevant area.
        - k: The (maximum) number of results to return.
    '''
    items = Shape.areatree.nearest((lat0, lon0, lat1, lon1), k)
    return [Shape.areas_by_id[i.id] for i in items]


class Shape:
    '''
        Base class of BlueSky shapes
    '''
    # Global counter to keep track of used shape ids
    max_area_id = 0

    # Weak-value dictionary of all Shape-derived objects by name, and id
    areas_by_id = WeakValueDictionary()
    areas_by_name = WeakValueDictionary()

    # RTree of all areas for efficient geospatial searching
    areatree = Index()

    @classmethod
    def reset(cls):
        ''' Reset shape data when simulation is reset. '''
        # Weak dicts and areatree should be cleared automatically
        # Reset max area id
        cls.max_area_id = 0

    def __init__(self, name, coordinates, top=1e9, bottom=-1e9):
        self.name = name
        self.coordinates = coordinates
        self.top = np.maximum(bottom, top)
        self.bottom = np.minimum(bottom, top)
        lat = coordinates[::2]
        lon = coordinates[1::2]
        self.bbox = [min(lat), min(lon), max(lat), max(lon)]

        # Global weak reference and tree storage
        self.area_id = Shape.max_area_id
        Shape.max_area_id += 1
        Shape.areas_by_id[self.area_id] = self
        Shape.areas_by_name[self.name] = self
        Shape.areatree.insert(self.area_id, self.bbox)

    def __del__(self):
        # Objects are removed automatically from the weak-value dicts,
        # but need to be manually removed from the rtree
        Shape.areatree.delete(self.area_id, self.bbox)

    @property
    def raw(self):
        ret = dict(name=self.name, shape=self.kind(), coordinates=self.coordinates)
        if hasattr(self, 'color'):
            ret['color'] = self.color
        return ret

    def checkInside(self, lat, lon, alt):
        ''' Returns True (or boolean array) if coordinate lat, lon, alt lies
            within this shape.

            Reimplement this function in the derived shape classes for this to
            work.
        '''
        return False

    def _str_vrange(self):
        if self.top < 9e8:
            if self.bottom > -9e8:
                return f' with altitude between {self.bottom} and {self.top}'
            else:
                return f' with altitude below {self.top}'
        if self.bottom > -9e8:
            return f' with altitude above {self.bottom}'
        return ''

    def __str__(self):
        return f'{self.name} is a {self.raw["shape"]} with coordinates ' + \
            ', '.join(str(c) for c in self.coordinates) + self._str_vrange()

    @classmethod
    def kind(cls):
        ''' Return a string describing what kind of shape this is. '''
        return cls.__name__.upper()


class Line(Shape):
    ''' A line shape '''
    def __init__(self, name, coordinates):
        super().__init__(name, coordinates)

    def __str__(self):
        return f'{self.name} is a LINE with ' \
            f'start point ({self.coordinates[0]}, {self.coordinates[1]}), ' \
            f'and end point ({self.coordinates[2]}, {self.coordinates[3]}).'


class Box(Shape):
    ''' A box shape '''
    def __init__(self, name, coordinates, top=1e9, bottom=-1e9):
        super().__init__(name, coordinates, top, bottom)
        # Sort the order of the corner points
        self.lat0 = min(coordinates[0], coordinates[2])
        self.lon0 = min(coordinates[1], coordinates[3])
        self.lat1 = max(coordinates[0], coordinates[2])
        self.lon1 = max(coordinates[1], coordinates[3])

    def checkInside(self, lat, lon, alt):
        return ((self.lat0 <=  lat) & ( lat <= self.lat1)) & \
               ((self.lon0 <= lon) & (lon <= self.lon1)) & \
               ((self.bottom <= alt) & (alt <= self.top))


class Circle(Shape):
    ''' A circle shape '''
    def __init__(self, name, coordinates, top=1e9, bottom=-1e9):
        super().__init__(name, coordinates, top, bottom)
        self.clat   = coordinates[0]
        self.clon   = coordinates[1]
        self.r      = coordinates[2]

    def checkInside(self, lat, lon, alt):
        distance = kwikdist(self.clat, self.clon, lat, lon)  # [NM]
        inside   = (distance <= self.r) & (self.bottom <= alt) & (alt <= self.top)
        return inside

    def __str__(self):
        return f'{self.name} is a CIRCLE with ' \
            f'center ({self.clat}, {self.clon}) ' \
            f'and radius {self.r}.' + self._str_vrange()


class Poly(Shape):
    ''' A polygon shape '''
    def __init__(self, name, coordinates, top=1e9, bottom=-1e9):
        super().__init__(name, coordinates, top, bottom)
        self.border = Path(np.reshape(coordinates, (len(coordinates) // 2, 2)))

    def checkInside(self, lat, lon, alt):
        points = np.vstack((lat,lon)).T
        inside = np.all((self.border.contains_points(points), self.bottom <= alt, alt <= self.top), axis=0)
        return inside



''' BlueSky OpenGL line and polygon (areafilter) drawing. '''
import numpy as np
from bluesky.core import Signal
from bluesky.network import context as ctx
from bluesky.network.subscriber import subscriber
from bluesky.stack import command
from bluesky.ui import palette
from bluesky.ui.polytools import PolygonSet
from bluesky.ui.qtgl import console
from bluesky.ui.qtgl import glhelpers as glh
from bluesky.network.sharedstate import ActData
from bluesky.network.common import ActionType


palette.set_default_colours(
    polys=(0, 0, 255),
    previewpoly=(0, 204, 255)
) 

# Static defines
POLYPREV_SIZE = 500 # Max number of vertices in preview off new edited polygon
POLY_SIZE = 20000   # Max total number of vertices when summing all polygon vertices


class Poly(glh.RenderObject, layer=-20):
    ''' Poly OpenGL object. '''

    # Per remote node attributes
    show_poly: ActData[int] = ActData(1)
    polys: ActData[dict] = ActData(group='poly')
    bufdata: ActData[dict] = ActData()

    @command
    def showpoly(self, flag:int=None):
        ''' Toggle drawing of polygon shapes between off, outline, and outline+fill. '''
        # Cycle aircraft label through detail level 0,1,2
        if flag is None:
            self.show_poly = (self.show_poly + 1) % 3

        # Or use the argument if it is an integer
        else:
            self.show_poly = min(2,max(0,flag))

    def __init__(self, parent=None):
        super().__init__(parent)
        # Polygon preview object
        self.polyprev = glh.VertexArrayObject(glh.gl.GL_LINE_LOOP)

        # Fixed polygons
        self.allpolys = glh.VertexArrayObject(glh.gl.GL_LINES)
        self.allpfill = glh.VertexArrayObject(glh.gl.GL_TRIANGLES)

        self.prevmousepos = (0, 0)

        Signal('cmdline_stacked').connect(self.cmdline_stacked)
        Signal('radarmouse').connect(self.previewpoly)

    def create(self):
        # self.polyprev.create(vertex=POLYPREV_SIZE * 8,
        #                      color=palette.previewpoly, usage=glh.gl.GL_DYNAMIC_DRAW)
        self.polyprev.create(vertex=POLYPREV_SIZE * 8,
                             color=palette.previewpoly, usage=glh.GLBuffer.UsagePattern.DynamicDraw)
        self.allpolys.create(vertex=POLY_SIZE * 16, color=POLY_SIZE * 8)
        self.allpfill.create(vertex=POLY_SIZE * 24,
                             color=np.append(palette.polys, 50))

    def draw(self):
        # Send the (possibly) updated global uniforms to the buffer
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_LATLON)

        # --- DRAW THE MAP AND COASTLINES ---------------------------------------------
        # Map and coastlines: don't wrap around in the shader
        self.shaderset.enable_wrap(False)

        # --- DRAW PREVIEW SHAPE (WHEN AVAILABLE) -----------------------------
        self.polyprev.draw()

        # --- DRAW CUSTOM SHAPES (WHEN AVAILABLE) -----------------------------
        if self.show_poly > 0:
            self.allpolys.draw()
            if self.show_poly > 1:
                self.allpfill.draw()
        
    def cmdline_stacked(self, cmd, args):
        if cmd in ['AREA', 'BOX', 'POLY', 'POLYGON', 'CIRCLE', 'LINE', 'POLYLINE']:
            self.polyprev.set_vertex_count(0)

    # def previewpoly(self, shape_type, data_in=None):
    def previewpoly(self, mouseevent):
        if mouseevent.type() != mouseevent.Type.MouseMove:
            return
        mousepos = (mouseevent.pos().x(), mouseevent.pos().y())
        # Check if we are updating a preview poly
        if mousepos != self.prevmousepos:
            cmd = console.get_cmd()
            nargs = len(console.get_args())
            if cmd in ['AREA', 'BOX', 'POLY', 'POLYLINE',
                        'POLYALT', 'POLYGON', 'CIRCLE', 'LINE'] and nargs >= 2:
                self.prevmousepos = mousepos
                try:
                    # get the largest even number of points
                    start = 0 if cmd == 'AREA' else 3 if cmd == 'POLYALT' else 1
                    end = ((nargs - start) // 2) * 2 + start
                    data = [float(v) for v in console.get_args()[start:end]]
                    data += self.glsurface.pixelCoordsToLatLon(*mousepos)
                    self.glsurface.makeCurrent()
                    if cmd is None:
                        self.polyprev.set_vertex_count(0)
                        return
                    if cmd in ['BOX', 'AREA']:
                        # For a box (an area is a box) we need to add two additional corners
                        polydata = np.zeros(8, dtype=np.float32)
                        polydata[0:2] = data[0:2]
                        polydata[2:4] = data[2], data[1]
                        polydata[4:6] = data[2:4]
                        polydata[6:8] = data[0], data[3]
                    else:
                        polydata = np.array(data, dtype=np.float32)

                    if cmd[-4:] == 'LINE':
                        self.polyprev.set_primitive_type(glh.gl.GL_LINE_STRIP)
                    else:
                        self.polyprev.set_primitive_type(glh.gl.GL_LINE_LOOP)

                    self.polyprev.update(vertex=polydata)


                except ValueError:
                    pass

    @subscriber(topic='POLY')
    def update_poly_data(self, data):
        if ctx.action in (ActionType.Reset, ActionType.ActChange):
            # Simulation reset: Clear all entries
            self.bufdata.clear()
            self.allpolys.set_vertex_count(0)
            self.allpfill.set_vertex_count(0)
            names = data.polys.keys()
        
        # The data argument passed to this subscriber contains all poly
        # data. We only need the current updates, which are in ctx.action_content
        elif ctx.action == ActionType.Delete:
            # Delete action contains a list of poly names to delete
            names = ctx.action_content['polys']
        else:
            # All other updates contain a dict with names as keys
            # and updated/new polys as items
            names = ctx.action_content['polys'].keys()

        # We're either updating a polygon, or deleting it.
        for name in names:
            # Always delete the old processed data
            self.bufdata.pop(name, None)

            if ctx.action != ctx.action.Delete:
                polydata = data.polys[name]
                try:
                    shape = polydata['shape']
                    coordinates = polydata['coordinates']
                    color = polydata.get('color', palette.polys)
                    self.bufdata[name] = self.genbuffers(shape, coordinates, color)
                except:
                    print("Could not process incoming poly data")
                
        if self.bufdata:
            self.glsurface.makeCurrent()
            contours, fills, colors = zip(*self.bufdata.values())
            # Create contour buffer with color
            self.allpolys.update(vertex=np.concatenate(contours),
                                    color=np.concatenate(colors))

            # Create fill buffer
            self.allpfill.update(vertex=np.concatenate(fills))
        else:
            self.allpolys.set_vertex_count(0)
            self.allpfill.set_vertex_count(0)

    @staticmethod
    def genbuffers(shape, coordinates, color=None):
        ''' Generate outline, fill, and colour buffers for given shape. '''
        # Break up polyline list of (lat,lon)s into separate line segments
        if shape == 'LINE' or shape[:4] == 'POLY':
            # Input data is list or array: [lat0,lon0,lat1,lon1,lat2,lon2,lat3,lon3,..]
            newdata = np.array(coordinates, dtype=np.float32)

        elif shape == 'BOX':
            # Convert box coordinates into polyline list
            # BOX: 0 = lat0, 1 = lon0, 2 = lat1, 3 = lon1 , use bounding box
            newdata = np.array([coordinates[0], coordinates[1],
                            coordinates[0], coordinates[3],
                            coordinates[2], coordinates[3],
                            coordinates[2], coordinates[1]], dtype=np.float32)

        elif shape == 'CIRCLE':
            # Input data is latctr,lonctr,radius[nm]
            # Convert circle into polyline list

            # Circle parameters
            Rearth = 6371000.0 # radius of the Earth [m]
            numPoints = 72 # number of straight line segments that make up the circrle

            # Inputs
            lat0 = coordinates[0] # latitude of the center of the circle [deg]
            lon0 = coordinates[1] # longitude of the center of the circle [deg]
            Rcircle = coordinates[2] * 1852.0 # radius of circle [NM]

            # Compute flat Earth correction at the center of the experiment circle
            coslatinv = 1.0 / np.cos(np.deg2rad(lat0))

            # compute the x and y coordinates of the circle
            angles    = np.linspace(0.0, 2.0 * np.pi, numPoints)   # ,endpoint=True) # [rad]

            # Calculate the circle coordinates in lat/lon degrees.
            # Use flat-earth approximation to convert from cartesian to lat/lon.
            latCircle = lat0 + np.rad2deg(Rcircle * np.sin(angles) / Rearth)  # [deg]
            lonCircle = lon0 + np.rad2deg(Rcircle * np.cos(angles) * coslatinv / Rearth)  # [deg]

            # make the data array in the format needed to plot circle
            newdata = np.empty(2 * numPoints, dtype=np.float32)  # Create empty array
            newdata[0::2] = latCircle  # Fill array lat0,lon0,lat1,lon1....
            newdata[1::2] = lonCircle

        # Create polygon contour buffer
        # Distinguish between an open and a closed contour.
        # If this is a closed contour, add the first vertex again at the end
        # and add a fill shape
        if shape[-4:] == 'LINE':
            contourbuf = np.empty(2 * len(newdata) - 4, dtype=np.float32)
            contourbuf[0::4]   = newdata[0:-2:2]  # lat
            contourbuf[1::4]   = newdata[1:-2:2]  # lon
            contourbuf[2::4] = newdata[2::2]  # lat
            contourbuf[3::4] = newdata[3::2]  # lon
            fillbuf = np.array([], dtype=np.float32)
        else:
            contourbuf = np.empty(2 * len(newdata), dtype=np.float32)
            contourbuf[0::4]   = newdata[0::2]  # lat
            contourbuf[1::4]   = newdata[1::2]  # lon
            contourbuf[2:-2:4] = newdata[2::2]  # lat
            contourbuf[3:-3:4] = newdata[3::2]  # lon
            contourbuf[-2:]    = newdata[0:2]
            pset = PolygonSet()
            pset.addContour(newdata)
            fillbuf = np.array(pset.vbuf, dtype=np.float32)

        # Define color buffer for outline
        defclr = tuple(color or palette.polys) + (255,)
        colorbuf = np.array(len(contourbuf) // 2 * defclr, dtype=np.uint8)

        # Store new or updated polygon by name, and concatenated with the
        # other polys
        return contourbuf, fillbuf, colorbuf



from PyQt6.QtWidgets import QListView
from PyQt6.QtGui import QPalette


class StackList(QListView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setObjectName('StackList')
        self.setBackgroundRole(QPalette.ColorRole.NoRole)
        self.setAutoFillBackground(True)
        self.setStyleSheet('background-color: transparent')
    
    def keyPressEvent(self, event) -> None:
        print('stacklist')
        return super().keyPressEvent(event)


import openap
from openap import top

from bluesky import core, stack, traf


def init_plugin():
    """Plugin initialisation function."""
    optimization = Optimization()
    config = {
        "plugin_name": "OPTIMIZATION",
        "plugin_type": "sim",
    }
    return config


class Optimization(core.Entity):
    def __init__(self):
        super().__init__()

    @stack.command
    def optimalflight(
        self,
        acid: "acid",
        actype: str,
        lat0: float,
        lon0: float,
        lat1: float,
        lon1: float,
        m0: float = 0.85,
    ):
        """Optimize flight plan for aircraft 'acid' and add waypoints using CRE/ADDWPT.

        Parameters:
            acid: Aircraft ID
            actype: Aircraft type.
            origin_lat: Latitude for origin.
            origin_lon: Longitude for origin.
            dest_lat: Latitude for destination.
            dest_lon: Longitude for destination.
        """
        optimizer = top.CompleteFlight(
            actype, lat0, lon0, lat1, lon1, m0=m0, use_synonyms=True
        )
        flight = optimizer.trajectory()

        first = flight.iloc[0]
        stack.stack(
            f"CRE {traf.id[acid]},{actype},{first.latitude},{first.longitude},"
            f"{first.heading} {first.altitude} {first.tas}"
        )

        for _, point in flight.iloc[1:].iterrows():
            lat = point["latitude"]
            lon = point["longitude"]
            alt = point["altitude"]
            spd = point["tas"]
            stack.stack(f"ADDWPT {traf.id[acid]} {lat} {lon} {alt} {spd}")

        return True, "Optimized flight plan generated."

    @stack.command
    def optimalflightod(self, acid: "acid", actype: str, origin: str, destination: str):
        """
        Optimize flight plan for aircraft 'acid' using ICAO airpoint codes for origin and destination.
        Retrieves lat/lon from openap.nav.airport.
        """
        # Retrieve airport data
        origin_airport = openap.nav.airport(origin)
        dest_airport = openap.nav.airport(destination)
        lat0, lon0 = origin_airport["lat"], origin_airport["lon"]
        lat1, lon1 = dest_airport["lat"], dest_airport["lon"]

        return self.optimalflight(acid, actype, lat0, lon0, lat1, lon1)



""" BlueSky aircraft performance calculations using BADA 3.xx."""
import numpy as np
import bluesky as bs
from bluesky.tools.aero import kts, ft, g0, vtas2cas, vcas2tas
from bluesky.traffic.performance.perfbase import PerfBase
from bluesky.traffic.performance.legacy.performance import esf, phases, calclimits, PHASE
from bluesky import settings
from bluesky import stack
from . import coeff_bada


# Check if BADA is complete
coeff_bada.check()
# Register settings defaults
settings.set_variable_defaults(performance_dt=1.0, verbose=False)


class BADA(PerfBase):
    """
    Aircraft performance implementation based on BADA.
    Methods:

        reset()           : clear current database
        create(actype)    : initialize new aircraft with performance parameters
        delete(idx)       : remove performance parameters from deleted aircraft
        perf()            : calculate aircraft performance
        limits()          : calculate flight envelope

    Created by  : Isabel Metz
    Note: This class is based on
        EUROCONTROL. User Manual for the Base of Aircraft Data (BADA) Revision 3.12,
        EEC Technical/Scientific Report No. 14/04/24-44 edition, 2014.
    """
    def __init__(self):
        super().__init__()
        # Load coefficient files if we haven't yet
        coeff_bada.init()

        self.warned = False     # Flag: Did we warn for default perf parameters yet?
        self.warned2 = False    # Flag: Use of piston engine aircraft?

        # Flight performance scheduling
        self.warned2 = False        # Flag: Did we warn for default engine parameters yet?

        # Register the per-aircraft parameter arrays
        with self.settrafarrays():
            # engine
            self.jet        = np.array([])
            self.turbo      = np.array([])
            self.piston     = np.array([])
            # masses and dimensions
            self.mass       = np.array([])  # effective mass [kg]
            # self.mref = np.array([]) # ref. mass [kg]: 70% between min and max. mass
            self.mmin       = np.array([])  # OEW (or assumption) [kg]
            self.mmax       = np.array([])  # MTOW (or assumption) [kg]
            # self.mpyld = np.array([]) # MZFW-OEW (or assumption) [kg]
            self.gw         = np.array([])  # weight gradient on max. alt [m/kg]
            self.Sref       = np.array([])  # wing reference surface area [m^2]

            # flight enveloppe
            self.vmto       = np.array([])  # min TO spd [m/s]
            self.vmic       = np.array([])  # min climb spd [m/s]
            self.vmcr       = np.array([])  # min cruise spd [m/s]
            self.vmap       = np.array([])  # min approach spd [m/s]
            self.vmld       = np.array([])  # min landing spd [m/s]

            self.vmo        = np.array([])   # max operating speed [m/s]
            self.mmo        = np.array([])   # max operating mach number [-]
            self.hmax       = np.array([])   # max. alt above standard MSL (ISA) at MTOW [m]
            self.hmaxact    = np.array([])   # max. alt depending on temperature gradient [m]
            self.hmo        = np.array([])   # max. operating alt abov standard MSL [m]
            self.gt         = np.array([])   # temp. gradient on max. alt [ft/k]
            self.maxthr     = np.array([])   # maximum thrust [N]

            # Buffet Coefficients
            self.clbo       = np.array([])    # buffet onset lift coefficient [-]
            self.k          = np.array([])    # buffet coefficient [-]
            self.cm16       = np.array([])    # CM16

            # reference CAS speeds
            self.cascl      = np.array([])    # climb [m/s]
            self.cascr      = np.array([])    # cruise [m/s]
            self.casdes     = np.array([])    # descent [m/s]

            #reference mach numbers [-]
            self.macl       = np.array([])    # climb
            self.macr       = np.array([])    # cruise
            self.mades      = np.array([])    # descent

            # parasitic drag coefficients per phase [-]
            self.cd0to      = np.array([])    # phase takeoff
            self.cd0ic      = np.array([])    # phase initial climb
            self.cd0cr      = np.array([])    # phase cruise
            self.cd0ap      = np.array([])    # phase approach
            self.cd0ld      = np.array([])    # phase land
            self.gear       = np.array([])    # drag due to gear down

            # induced drag coefficients per phase [-]
            self.cd2to      = np.array([])    # phase takeoff
            self.cd2ic      = np.array([])    # phase initial climb
            self.cd2cr      = np.array([])    # phase cruise
            self.cd2ap      = np.array([])    # phase approach
            self.cd2ld      = np.array([])    # phase land

            # max climb thrust coefficients
            self.ctcth1      = np.array([])   # jet/piston [N], turboprop [ktN]
            self.ctcth2      = np.array([])   # [ft]
            self.ctcth3      = np.array([])   # jet [1/ft^2], turboprop [N], piston [ktN]

            # reduced climb power coefficient
            self.cred       = np.array([])    # [-]

            # 1st and 2nd thrust temp coefficient
            self.ctct1      = np.array([])    # [k]
            self.ctct2      = np.array([])    # [1/k]
            self.dtemp      = np.array([])    # [k]

            # Descent Thrust Coefficients
            # Note: Ctdes,app and Ctdes,lnd assume a 3 degree descent gradient during app and lnd
            self.ctdesl      = np.array([])   # low alt descent thrust coefficient [-]
            self.ctdesh      = np.array([])   # high alt descent thrust coefficient [-]
            self.ctdesa      = np.array([])   # approach thrust coefficient [-]
            self.ctdesld     = np.array([])   # landing thrust coefficient [-]

            # transition altitude for calculation of descent thrust
            self.hpdes       = np.array([])   # [m]

            # Energy Share Factor
            self.ESF         = np.array([])   # [-]

            # reference speed during descent
            self.vdes        = np.array([])   # [m/s]
            self.mdes        = np.array([])   # [-]

            # flight phase
            self.phase       = np.array([])
            self.post_flight = np.array([])   # taxi prior of post flight?
            self.pf_flag     = np.array([])

            # Thrust specific fuel consumption coefficients
            self.cf1         = np.array([])   # jet [kg/(min*kN)], turboprop [kg/(min*kN*knot)], piston [kg/min]
            self.cf2         = np.array([])   # [knots]
            self.cf3         = np.array([])   # [kg/min]
            self.cf4         = np.array([])   # [ft]
            self.cf_cruise   = np.array([])   # [-]

            # performance
            self.thrust      = np.array([])   # thrust
            self.D           = np.array([])   # drag
            self.fuelflow    = np.array([])   # fuel flow

            # ground
            self.tol         = np.array([])   # take-off length[m]
            self.ldl         = np.array([])   # landing length[m]
            self.ws          = np.array([])   # wingspan [m]
            self.len         = np.array([])   # aircraft length[m]
            self.gr_acc      = np.array([])   # ground acceleration [m/s^2]

            # limit settings
            self.limspd      = np.array([])  # limit speed
            self.limspd_flag = np.array([], dtype=bool)  # flag for limit spd - we have to test for max and min
            self.limalt      = np.array([])  # limit altitude
            self.limalt_flag = np.array([])  # A need to limit altitude has been detected
            self.limvs       = np.array([])  # limit vertical speed due to thrust limitation
            self.limvs_flag  = np.array([])  # A need to limit V/S detected

    def engchange(self, acid, engid=None):
        return False, "BADA performance model doesn't allow changing engine type"

    def create(self, n=1):
        super().create(n)
        """CREATE NEW AIRCRAFT"""
        actypes = bs.traf.type[-n:]

        # note: coefficients are initialized in SI units

        # general
        # designate aircraft to its aircraft type
        for i, actype in enumerate(actypes):
            syn, coeff = coeff_bada.getCoefficients(actype)
            if syn:
                continue
            syn, coeff = coeff_bada.getCoefficients('B744')
            bs.traf.type[-n + i] = syn.accode

            if not settings.verbose:
                if not self.warned:
                    print("Aircraft is using default B747-400 performance.")
                    self.warned = True
            else:
                print("Flight " + bs.traf.id[-n:] + " has an unknown aircraft type, " + actype + ", BlueSky then uses default B747-400 performance.")

        # designate aicraft to its aircraft type
        self.jet[-n:]       = 1 if coeff.engtype == 'Jet' else 0
        self.turbo[-n:]     = 1 if coeff.engtype == 'Turboprop' else 0
        self.piston[-n:]    = 1 if coeff.engtype == 'Piston' else 0

        # Initial aircraft mass is currently reference mass.
        # BADA 3.12 also supports masses between 1.2*mmin and mmax
        self.mass[-n:]      = coeff.m_ref * 1000.0
        self.mmin[-n:]      = coeff.m_min * 1000.0
        self.mmax[-n:]      = coeff.m_max * 1000.0

        # self.mpyld = np.append(self.mpyld, coeff.mpyld[coeffidx]*1000)
        self.gw[-n:]        = coeff.mass_grad * ft

        # Surface Area [m^2]
        self.Sref[-n:]      = coeff.S

        # flight envelope
        # minimum speeds per phase
        self.vmto[-n:]      = coeff.Vstall_to * coeff.CVmin_to * kts
        self.vmic[-n:]      = coeff.Vstall_ic * coeff.CVmin * kts
        self.vmcr[-n:]      = coeff.Vstall_cr * coeff.CVmin * kts
        self.vmap[-n:]      = coeff.Vstall_ap * coeff.CVmin * kts
        self.vmld[-n:]      = coeff.Vstall_ld * coeff.CVmin * kts
        self.vmin[-n:]      = 0.0
        self.vmo[-n:]       = coeff.VMO * kts
        self.mmo[-n:]       = coeff.MMO
        self.vmax[-n:]      = self.vmo[-n:]

        # max. altitude parameters
        self.hmo[-n:]       = coeff.h_MO * ft
        self.hmax[-n:]      = coeff.h_max * ft
        self.hmaxact[-n:]   = coeff.h_max * ft  # initialize with hmax
        self.gt[-n:]        = coeff.temp_grad * ft

        # max thrust setting
        self.maxthr[-n:]    = 1e6  # initialize with excessive setting to avoid unrealistic limit setting

        # Buffet Coefficients
        self.clbo[-n:]      = coeff.Clbo
        self.k[-n:]         = coeff.k
        self.cm16[-n:]      = coeff.CM16

        # reference speeds
        # reference CAS speeds
        self.cascl[-n:]     = coeff.CAScl1[0] * kts
        self.cascr[-n:]     = coeff.CAScr1[0] * kts
        self.casdes[-n:]    = coeff.CASdes1[0] * kts

        # reference mach numbers
        self.macl[-n:]      = coeff.Mcl[0]
        self.macr[-n:]      = coeff.Mcr[0]
        self.mades[-n:]     = coeff.Mdes[0]

        # reference speed during descent
        self.vdes[-n:]      = coeff.Vdes_ref * kts
        self.mdes[-n:]      = coeff.Mdes_ref

        # aerodynamics
        # parasitic drag coefficients per phase
        self.cd0to[-n:]     = coeff.CD0_to
        self.cd0ic[-n:]     = coeff.CD0_ic
        self.cd0cr[-n:]     = coeff.CD0_cr
        self.cd0ap[-n:]     = coeff.CD0_ap
        self.cd0ld[-n:]     = coeff.CD0_ld
        self.gear[-n:]      = coeff.CD0_gear

        # induced drag coefficients per phase
        self.cd2to[-n:]     = coeff.CD2_to
        self.cd2ic[-n:]     = coeff.CD2_ic
        self.cd2cr[-n:]     = coeff.CD2_cr
        self.cd2ap[-n:]     = coeff.CD2_ap
        self.cd2ld[-n:]     = coeff.CD2_ld

        # reduced climb coefficient
        self.cred[-n:] = np.where(
            self.jet[-n:], coeff.Cred_jet,
            np.where(self.turbo[-n:], coeff.Cred_turboprop, coeff.Cred_piston)
        )

        # commented due to vectrization
        # # NOTE: model only validated for jet and turbo aircraft
        # if self.piston[-n:] and not self.warned2:
        #     print "Using piston aircraft performance.",
        #     print "Not valid for real performance calculations."
        #     self.warned2 = True

        # performance

        # max climb thrust coefficients
        self.ctcth1[-n:]    = coeff.CTC[0]  # jet/piston [N], turboprop [ktN]
        self.ctcth2[-n:]    = coeff.CTC[1]  # [ft]
        self.ctcth3[-n:]    = coeff.CTC[2]  # jet [1/ft^2], turboprop [N], piston [ktN]

        # 1st and 2nd thrust temp coefficient
        self.ctct1[-n:]     = coeff.CTC[3]  # [k]
        self.ctct2[-n:]     = coeff.CTC[4]  # [1/k]
        self.dtemp[-n:]     = 0.0  # [k], difference from current to ISA temperature. At the moment: 0, as ISA environment

        # Descent Fuel Flow Coefficients
        # Note: Ctdes,app and Ctdes,lnd assume a 3 degree descent gradient during app and lnd
        self.ctdesl[-n:]    = coeff.CTdes_low
        self.ctdesh[-n:]    = coeff.CTdes_high
        self.ctdesa[-n:]    = coeff.CTdes_app
        self.ctdesld[-n:]   = coeff.CTdes_land

        # transition altitude for calculation of descent thrust
        self.hpdes[-n:]     = coeff.Hp_des * ft
        self.ESF[-n:]       = 1.0  # neutral initialisation

        # flight phase
        self.phase[-n:]       = PHASE["None"]
        self.post_flight[-n:] = False  # we assume prior
        self.pf_flag[-n:]     = True

        # Thrust specific fuel consumption coefficients
        # prevent from division per zero in fuelflow calculation
        self.cf1[-n:]       = coeff.Cf1
        self.cf2[-n:]       = 1.0 if coeff.Cf2 < 1e-9 else coeff.Cf2
        self.cf3[-n:]       = coeff.Cf3
        self.cf4[-n:]       = 1.0 if coeff.Cf4 < 1e-9 else coeff.Cf4
        self.cf_cruise[-n:] = coeff.Cf_cruise

        self.thrust[-n:] = 0.0
        self.D[-n:]         = 0.0
        self.fuelflow[-n:]  = 0.0

        # ground
        self.tol[-n:]       = coeff.TOL
        self.ldl[-n:]       = coeff.LDL
        self.ws[-n:]        = coeff.wingspan
        self.len[-n:]       = coeff.length
        # for now, BADA aircraft have the same acceleration as deceleration
        self.gr_acc[-n:]    = coeff.gr_acc

    def update(self, dt):
        ''' Periodic update function for performance calculations. '''
        # BADA version
        swbada = True
        delalt = bs.traf.selalt - bs.traf.alt
        # flight phase
        self.phase, self.bank = phases(bs.traf.alt, bs.traf.gs, delalt,
            bs.traf.cas, self.vmto, self.vmic, self.vmap, self.vmcr, self.vmld,
            bs.traf.ap.bankdef, bs.traf.bphase, bs.traf.swhdgsel, swbada)

        # AERODYNAMICS
        # Lift
        qS = 0.5*bs.traf.rho*np.maximum(1.,bs.traf.tas)*np.maximum(1.,bs.traf.tas)*self.Sref
        cl = self.mass*g0/(qS*np.cos(self.bank))*(self.phase!=PHASE["GD"])+ 0.*(self.phase==PHASE["GD"])

        # Drag
        # Drag Coefficient

        # phases TO, IC, CR
        cdph = self.cd0cr+self.cd2cr*(cl*cl)

        # phase AP
        # in case approach coefficients in OPF-Files are set to zero:
        #Use cruise values instead
        cdapp = np.where(self.cd0ap !=0, self.cd0ap+self.cd2ap*(cl*cl), cdph)

        # phase LD
        # in case landing coefficients in OPF-Files are set to zero:
        #Use cruise values instead
        cdld = np.where(self.cd0ld !=0, self.cd0ld+self.cd2ld*(cl*cl), cdph)


        # now combine phases
        cd = (self.phase==PHASE['TO'])*cdph + (self.phase==PHASE["IC"])*cdph + (self.phase==PHASE["CR"])*cdph \
            + (self.phase==PHASE['AP'])*cdapp + (self.phase ==PHASE['LD'])*cdld

        # Drag:
        self.D = cd*qS

        # energy share factor and crossover altitude

        # conditions
        epsalt = np.array([0.001]*bs.traf.ntraf)
        climb = np.array(delalt > epsalt)
        descent = np.array(delalt<-epsalt)
        lvl = np.array(np.abs(delalt)<0.0001)*1

        # energy share factor
        delspd = bs.traf.aporasas.tas - bs.traf.tas
        selmach = bs.traf.selspd < 2.0
        self.ESF = esf(bs.traf.alt, bs.traf.M, climb, descent, delspd, selmach)

        # THRUST
        # 1. climb: max.climb thrust in ISA conditions (p. 32, BADA User Manual 3.12)
        # condition: delta altitude positive


        # temperature correction for non-ISA (as soon as applied)
        #            ThrISA = (1-self.ctct2*(self.dtemp-self.ctct1))
        # jet
        # condition
        cljet = np.logical_and.reduce([climb, self.jet]) * 1

        # thrust
        Tj = self.ctcth1* (1-(bs.traf.alt/ft)/self.ctcth2+self.ctcth3*(bs.traf.alt/ft)*(bs.traf.alt/ft))

        # combine jet and default aircraft
        Tjc = cljet*Tj # *ThrISA

        # turboprop
        # condition
        clturbo = np.logical_and.reduce([climb, self.turbo]) * 1

        # thrust
        Tt = self.ctcth1/np.maximum(1.,bs.traf.tas/kts)*(1-(bs.traf.alt/ft)/self.ctcth2)+self.ctcth3

        # merge
        Ttc = clturbo*Tt # *ThrISA

        # piston
        clpiston = np.logical_and.reduce([climb, self.piston])*1
        Tp = self.ctcth1*(1-(bs.traf.alt/ft)/self.ctcth2)+self.ctcth3/np.maximum(1.,bs.traf.tas/kts)
        Tpc = clpiston*Tp

        # max climb thrust for futher calculations (equals maximum avaliable thrust)
        maxthr = Tj*self.jet + Tt*self.turbo + Tp*self.piston

        # 2. level flight: Thr = D.
        Tlvl = lvl*self.D

        # 3. Descent: condition: vs negative/ H>hdes: fixed formula. H<hdes: phase cr, ap, ld

        # above or below Hpdes? Careful! If non-ISA: ALT must be replaced by Hp!
        delh = (bs.traf.alt - self.hpdes)

        # above Hpdes:
        high = np.array(delh>0)
        Tdesh = maxthr*self.ctdesh*np.logical_and.reduce([descent, high])

        # below Hpdes
        low = np.array(delh<0)
        # phase cruise
        Tdeslc = maxthr*self.ctdesl*np.logical_and.reduce([descent, low, (self.phase==PHASE['CR'])])
        # phase approach
        Tdesla = maxthr*self.ctdesa*np.logical_and.reduce([descent, low, (self.phase==PHASE['AP'])])
        # phase landing
        Tdesll = maxthr*self.ctdesld*np.logical_and.reduce([descent, low, (self.phase==PHASE['LD'])])
        # phase ground: minimum descent thrust as a first approach
        Tgd = np.minimum.reduce([Tdesh, Tdeslc])*(self.phase==PHASE['GD'])

        # merge all thrust conditions
        T = np.maximum.reduce([Tjc, Ttc, Tpc, Tlvl, Tdesh, Tdeslc, Tdesla, Tdesll, Tgd])


        # vertical speed
        # vertical speed. Note: ISA only ( tISA = 1 )
        # for climbs: reducing factor (reduced climb power) is multiplied
        # cred applies below 0.8*hmax and for climbing aircraft only
        hcred = np.array(bs.traf.alt < (self.hmaxact*0.8))
        clh = np.logical_and.reduce([hcred, climb])
        cred = self.cred*clh
        cpred = 1-cred*((self.mmax-self.mass)/(self.mmax-self.mmin))


        # switch for given vertical speed selvs
        if (bs.traf.selvs.any()>0.) or (bs.traf.selvs.any()<0.):
            # thrust = f(selvs)
            T_vs = ((bs.traf.selvs!=0) * \
                    (((bs.traf.aporasas.vs * np.sign(delalt)*self.mass*g0)/ \
                    (self.ESF*np.maximum(bs.traf.eps,bs.traf.tas)*cpred)) \
                    + self.D)) + ((bs.traf.selvs==0)*T)

            # limit minimum thrust in descent to idle thrust
            T = np.where(descent, np.maximum(T_vs, T), T)
        self.thrust = T


        # Fuel consumption
        # thrust specific fuel consumption - jet
        # thrust
        etaj = self.cf1*(1.0+(bs.traf.tas/kts)/self.cf2)
        # merge
        ej = etaj*self.jet

        # thrust specific fuel consumption - turboprop

        # thrust
        etat = self.cf1*(1.-(bs.traf.tas/kts)/self.cf2)*((bs.traf.tas/kts)/1000.)
        # merge
        et = etat*self.turbo

        # thrust specific fuel consumption for all aircraft
        # eta is given in [kg/(min*kN)] - convert to [kg/(min*N)]
        eta = np.maximum.reduce([ej, et])/1000.

        # nominal fuel flow - (jet & turbo) and piston
        # condition jet,turbo:
        jt = np.maximum.reduce([self.jet, self.turbo])
        pdf = np.maximum.reduce ([self.piston])

        fnomjt = eta*self.thrust*jt
        fnomp = self.cf1*pdf
        # merge
        fnom = fnomjt + fnomp

        # minimal fuel flow jet, turbo and piston
        fminjt = self.cf3*(1-(bs.traf.alt/ft)/self.cf4)*jt
        fminp = self.cf3*pdf
        #merge
        fmin = fminjt + fminp

        # cruise fuel flow jet, turbo and piston
        fcrjt = eta*self.thrust*self.cf_cruise*jt
        fcrp = self.cf1*self.cf_cruise*pdf
        #merge
        fcr = fcrjt + fcrp

        # approach/landing fuel flow
        fal = np.maximum(fnom, fmin)

        # designate each aircraft to its fuelflow
        # takeoff
        ffto = fnom*(self.phase==PHASE['TO'])

        # initial climb
        ffic = fnom*(self.phase==PHASE['IC'])/2

        # phase cruise and climb
        cc = np.logical_and.reduce([climb, (self.phase==PHASE['CR'])])*1
        ffcc = fnom*cc

        # cruise and level
        ffcrl = fcr*lvl

        # descent cruise configuration
        cd2 = np.logical_and.reduce ([descent, (self.phase==PHASE['CR'])])*1
        ffcd = cd2*fmin

        # approach
        ffap = fal*(self.phase==PHASE['AP'])

        # landing
        ffld = fal*(self.phase==PHASE['LD'])

        # ground
        ffgd = fmin*(self.phase==PHASE['GD'])

        # fuel flow for each condition
        self.fuelflow = np.maximum.reduce([ffto, ffic, ffcc, ffcrl, ffcd, ffap, ffld, ffgd])/60. # convert from kg/min to kg/sec

        # update mass
        self.mass -= self.fuelflow * dt # Use fuelflow in kg/min



        # for aircraft on the runway and taxiways we need to know, whether they
        # are prior or after their flight
        self.post_flight = np.where(descent, True, self.post_flight)

        # when landing, we would like to stop the aircraft.
        bs.traf.aporasas.tas = np.where((bs.traf.alt <0.5)*(self.post_flight)*self.pf_flag, 0.0, bs.traf.aporasas.tas)


        # otherwise taxiing will be impossible afterwards
        # pf_flag is released so post_flight flag is only triggered once

        self.pf_flag = np.where ((bs.traf.alt <0.5)*(self.post_flight), False, self.pf_flag)

        # define acceleration: aircraft taxiing and taking off use ground acceleration,
        # landing aircraft use ground deceleration, others use standard acceleration
        # --> BADA uses the same value for ground acceleration as for deceleration
        self.axmax = ((self.phase == PHASE['IC']) + (self.phase == PHASE['CR']) + (self.phase == PHASE['AP']) + (self.phase == PHASE['LD'])) * 0.5 \
            + ((self.phase == PHASE['TO']) + (self.phase == PHASE['GD'])*(1-self.post_flight)) * self.gr_acc  \
            + (self.phase == PHASE['GD']) * self.post_flight * self.gr_acc

    def limits(self, intent_v, intent_vs, intent_h, ax):
        """FLIGHT ENVELPOE"""
        # summarize minimum speeds - ac in ground mode might be pushing back
        self.vmin =  (self.phase == 1) * self.vmto + (self.phase == 2) * self.vmic + (self.phase == 3) * self.vmcr + \
        (self.phase == 4) * self.vmap + (self.phase == 5) * self.vmld + (self.phase == 6) * -10.

        # maximum altitude: hmax/act = MIN[hmo, hmax+gt*(dtemp-ctc1)+gw*(mmax-mact)]
        #                   or hmo if hmx ==0 ()
        # at the moment just ISA atmosphere, dtemp  = 0
        c1 = self.dtemp - self.ctct1

        # if c1<0: c1 = 0
        # values above 0 remain, values below are replaced through 0
        c1m = np.array(c1<0)*0.00000000000001
        c1def = np.maximum(c1, c1m)

        self.hact = self.hmax+self.gt*c1def+self.gw*(self.mmax-self.mass)
        # if hmax in OPF File ==0: hmaxact = hmo, else minimum(hmo, hmact)
        self.hmaxact = (self.hmax==0)*self.hmo +(self.hmax !=0)*np.minimum(self.hmo, self.hact)

        # forwarding to tools
        self.limspd, self.limspd_flag, self.limalt, \
            self.limalt_flag, self.limvs, self.limvs_flag = calclimits(
                vtas2cas(intent_v, bs.traf.alt), bs.traf.gs,
                self.vmto, self.vmin, self.vmo, self.mmo,
                bs.traf.M, bs.traf.alt, self.hmaxact,
                intent_h, intent_vs, self.maxthr,
                self.thrust, self.D, bs.traf.tas,
                self.mass, self.ESF, self.phase)

        # Update desired sates with values within the flight envelope
        # When CAS is limited, it needs to be converted to TAS as only this TAS is used later on!
        allowed_tas = np.where(self.limspd_flag, vcas2tas(
            self.limspd, bs.traf.alt), intent_v)

        # Autopilot selected altitude [m]
        allowed_alt = np.where(self.limalt_flag, self.limalt, intent_h)

        # Autopilot selected vertical speed (V/S)
        allowed_vs = np.where(self.limvs_flag, self.limvs, intent_vs)

        return allowed_tas, allowed_vs, allowed_alt

    def show_performance(self, acid):
        # PERF acid command
        return True, \
            f"Flight phase: {self.phase[acid]}\n" + \
            f"Thrust: {self.thrust[acid] / 1000} kN\n" + \
            f"Drag: {self.D[acid] / 1000:.0f} kN\n" + \
            f"Fuel flow: {self.fuelflow[acid]:.2f} kg/s\n" + \
            f"Speed envelope: Min: {self.vmin[acid] / kts:.0f} MMO: {self.vmo[acid] / kts:.0f} kts M {self.mmo[acid]:.2f}\n" + \
            f"Ceiling: {self.hmax[acid] / ft:.0f} ft"



''' BlueSky Client class. '''
from collections import defaultdict

from bluesky.core.signal import Signal
from bluesky.stack.clientstack import process
from bluesky.network.node import Node
from bluesky.network.common import genid, GROUPID_CLIENT, GROUPID_SIM, GROUPID_DEFAULT

class Client(Node):
    def __init__(self, group_id=GROUPID_CLIENT):
        super().__init__(group_id)
        self.acttopics = defaultdict(set)
        self.discovery = None

        # Signals
        self.actnode_changed = Signal('actnode-changed')
        self.node_added.connect(self.actnode)

    def update(self):
        ''' Client periodic update function.

            Periodically call this function to allow client to receive and process data.
        '''
        super().update()

        # Process any waiting stacked commands
        process()
    
    def actnode(self, newact=None):
        ''' Set the new active node, or return the current active node. '''
        if newact:
            if newact not in self.nodes:
                print('Error selecting active node (unknown node)')
                return None
            if self.act_id is None:
                # This is the first time an active node is selected
                # We can now unsubscribe this method from the node-added signal
                self.node_added.disconnect(self.actnode)

            # Unsubscribe from previous node, subscribe to new one.
            if newact != self.act_id:
                for topic, groupset in self.acttopics.items():
                    for to_group in groupset:
                        if self.act_id:
                            self._unsubscribe(topic, self.act_id, to_group)
                        self._subscribe(topic, newact, to_group)
                self.act_id = newact
                self.actnode_changed.emit(newact)

        return self.act_id

    def _subscribe(self, topic, from_group=GROUPID_DEFAULT, to_group='', actonly=False):
        if from_group == GROUPID_DEFAULT:
            from_group = GROUPID_SIM
            if actonly:
                self.acttopics[topic].add(to_group)
                if self.act_id is not None:
                    from_group = self.act_id
                else:
                    return
        return super()._subscribe(topic, from_group, to_group)
    
    def _unsubscribe(self, topic, from_group=GROUPID_DEFAULT, to_group=''):
        if from_group == GROUPID_DEFAULT:
            from_group = GROUPID_SIM
            if topic in self.acttopics:
                self.acttopics[topic].discard(to_group)
                if self.act_id is not None:
                    from_group = self.act_id
                else:
                    return
        return super()._unsubscribe(topic, from_group, to_group)

    def addnodes(self, count=1, *node_ids, server_id=None):
        ''' Tell the specified server to add 'count' nodes. 
        
            If provided, create these nodes with the specified node ids.

            If no server_id is specified, the corresponding server of the
            currently-active node is targeted.
        '''
        self.send('ADDNODES', dict(count=count, node_ids=node_ids), server_id or genid(self.act_id[:-1], seqidx=0))



from matplotlib.path import Path
import json


try:
    from rtree.index import Index
except ImportError:
    print('Geofence plugin needs rtree.')
    class Index: ...
try:
    from shapely.geometry import Point
    from shapely.ops import nearest_points
except ImportError:
    print('Geofence plugin needs shapely.')

import bluesky as bs
from bluesky import stack
from bluesky.tools import aero, areafilter, geo

bs.settings.set_variable_defaults(geofence_dtlookahead=30)

def init_plugin():
    # Configuration parameters
    config = {
        'plugin_name': 'GEOFENCE',
        'plugin_type': 'sim',
        'reset': Geofence.reset
    }
    return config

@stack.command()
def geofence(name: 'txt', top: float = 999999, bottom: float = -999999, *coordinates: float):
    ''' Create a new geofence from the stack. 
        
        Arguments:
        - name: The name of the new geofence
        - top: The top of the geofence in feet.
        - bottom: The bottom of the geofence in feet.
        - coordinates: three or more lat/lon coordinates in degrees.
    '''
    Geofence(name, coordinates, top, bottom)
    return True, f'Created geofence {name}'


@stack.command()
def delgeofence(name: 'txt'):
    ''' Delete a geofence.'''
    Geofence.delete(name)

@stack.command()
def savegeofences(filename: 'txt'):
    ''' Save the current loaded geofences.'''
    if filename[-5:] != '.json':
        filename = filename + '.json'
    
    filepath = bs.resource(bs.settings.data_path) / 'geofences' / filename

    with open(filepath, 'w') as f:
        json.dump(Geofence.geo_save_dict, f, indent=4)

    stack.echo(f'Geofences saved to {filename}.')

@stack.command()
def loadgeofences(filename: 'txt'):
    '''Load a geofence file.'''
    if filename[-5:] != '.json':
        filename = filename + '.json'

    filepath = bs.resource(bs.settings.data_path) / 'geofences' / filename

    with open(filepath, 'r') as f:
        try:
            loaded_geo_dict = json.loads(f.read())
        except FileNotFoundError:
            stack.echo(f'File empty or does not exist.')
            return
        
    for geofence in loaded_geo_dict.values():
        Geofence(geofence['name'], geofence['coordinates'], geofence['top'], geofence['bottom'])

    stack.echo(f'Geofences loaded from {filename}.')

@stack.command()
def loadgeojson(filename: 'txt', name_col: 'txt'='name', top_col: 'txt'='top', bottom_col: 'txt'='bottom'):
    '''Load geofences from a GeoJSON file. Must be in EPSG:4326 format.'''
    try:
        import geopandas as gpd
        from fiona.errors import DriverError
    except ImportError:
        stack.echo('Geopandas is not installed. Cannot load geofences from geojson.')
        return

    # Load the geojson file into a geopandas dataframe
    if filename[-5:] != '.geojson':
        filename = filename + '.geojson'
    
    filepath = bs.resource(bs.settings.data_path) / 'geofences' / filename

    with open(filepath, 'r') as f:
        try:
            loaded_geo_dict = json.loads(f.read())
        except FileNotFoundError:
            stack.echo(f'File empty or does not exist.')

    # read the geojson dictionary into

    for geo_dict in loaded_geo_dict['features']:
        poly_type, [coords, *_] = geo_dict['geometry'].values()

        # only support polygons
        if not poly_type == 'Polygon':
            continue

        # get the name, top and bottom 
        geo_props = geo_dict['properties'] 

        
        # lowercase the name if it is not in the dictionary
        name_col = name_col.lower() if name_col not in geo_props.items() else name_col
        top_col = top_col.lower() if top_col not in geo_props.items() else top_col
        bottom_col = bottom_col.lower() if bottom_col not in geo_props.items() else bottom_col

        # if key is not in dictionary, skip
        try: 
            name = geo_props[name_col]
            top = geo_props[top_col]
            bottom = geo_props[bottom_col]

        except KeyError:
            stack.echo(f'Geofence has no {name_col}, {top_col} or {bottom_col} columns.')
            break
        
        # get coordinates into lat1,lon1,lat2,lon2,lat3,lon3...
        coordinates = [coord for xycoord in coords for coord in xycoord[::-1]]

        Geofence(name, coordinates, top, bottom)

    stack.echo(f'Geofences loaded from {filename}.')

@stack.command()
def drawgeofence(toggle: 'txt'='on'):
    '''Draw all geofences.'''
    if toggle.lower() == 'on':
        for name, geo_obj in Geofence.geo_by_name.items():
            bs.scr.objappend("POLY", name, geo_obj.coordinates)
    elif toggle.lower() == 'off':
        for name, geo_obj in Geofence.geo_by_name.items():
            bs.scr.objappend("POLY", name, None)

class Geofence(areafilter.Poly):
    ''' BlueSky Geofence class.
    
        This class subclasses Shape, and adds Geofence-specific data and methods.
    '''
    # Keep dicts of geofences by either name or rtree ID
    geo_by_name = dict()
    geo_by_id = dict()
    geo_name2id = dict()

    # Also have a dictionary used for saving and loading geofences
    geo_save_dict = dict()

    # Keep an Rtree of geofences
    geo_tree = Index()

    # Keep track of the geofences themselves that aircraft are hitting or intruding in
    # "intrusions" contains aircraft that are currently intruding inside a geofence, and a list
    # of the geofences they are intruding in
    # "hits" contains the geofences that aircraft are about to hit (or are intruding)
    intrusions = dict()
    hits = dict()
    
    # Unique intrusions dictionary: for each aircraft, keep track of the unique intrusions
    # that ever happened.
    unique_intrusions = dict()


    def __init__(self, name, coordinates, top=999999, bottom=-999999):
        super().__init__(name, coordinates, top=top, bottom=bottom)
        self.active = True
        #Add info to geofence save dictionary
        geo_dict = dict()
        geo_dict['name'] = name
        geo_dict['coordinates'] = coordinates
        geo_dict['top'] = top
        geo_dict['bottom'] = bottom
        Geofence.geo_save_dict[name] = geo_dict

        # Also add the class instance itself to the other dictionaries
        Geofence.geo_by_name[name] = self
        Geofence.geo_by_id[self.area_id] = self
        Geofence.geo_name2id[name] = self.area_id

        # Insert the geofence in the geofence Rtree
        Geofence.geo_tree.insert(self.area_id, self.bbox)

    def intersects(self, line):
        ''' Check whether given line intersects with this geofence poly. '''
        line_path = Path(line)
        return self.border.intersects_path(line_path)

    @classmethod
    def reset(cls):
        ''' Reset geofence database when simulation is reset. '''
        cls.geo_by_name.clear()
        cls.geo_by_id.clear()
        cls.geo_name2id.clear()
        cls.geo_save_dict.clear()
        cls.geo_tree = Index()
        cls.hits.clear()
        cls.intrusions.clear()
        cls.unique_intrusions.clear()

    @classmethod
    def delete(cls, name):
        geo_to_delete = cls.geo_by_name[name]
        cls.geo_by_name.pop(name)
        cls.geo_save_dict.pop(name)
        geo_id = cls.geo_name2id[name]
        cls.geo_tree.delete(geo_id, geo_to_delete.bbox)
        cls.geo_by_id.pop(geo_id)
        cls.geo_name2id.pop(name)

    def __del__(self):
        ...

    @classmethod
    def intersecting(cls, coordinates):
        '''Get the geofences that intersect coordinates (either bbox or point).'''
        poly_ids = list(cls.geo_tree.intersection(coordinates))
        return [cls.geo_by_id[id] for id in poly_ids], poly_ids

    @classmethod
    def detect_all(cls, traf, dtlookahead=None):
        if dtlookahead is None:
            dtlookahead = bs.settings.geofence_dtlookahead
        # Reset the hits dict
        cls.hits.clear()
        # Linearly extrapolate current state to prefict future position
        pred_lat, pred_lon = geo.kwikpos(traf.lat, traf.lon, traf.hdg, traf.gs / aero.nm * dtlookahead)
        for idx, line in enumerate(zip(traf.lat, traf.lon, pred_lat, pred_lon)):
            acid = traf.id[idx]
            # First a course detection based on geofence bounding boxes
            potential_hits= areafilter.get_intersecting(*line)
            # Then a fine-grained intersection detection
            hits = []
            for geofence in potential_hits:
                if geofence.intersects(line):
                    hits.append(geofence)
            cls.hits[acid] = hits
        return

    @classmethod
    def detect_inside(cls, traf):
        for idx, point in enumerate(zip(traf.lat, traf.lon, traf.alt)):
            acid = traf.id[idx]
            # First, a course detection based on geofence bounding boxes
            potential_intrusions, geo_ids = cls.intersecting([point[0], point[1]])
            ac_alt = point[2]/aero.ft
           
            # Then a fine-grained intrusion detection
            for i, geofence in enumerate(potential_intrusions):
                if ac_alt < geofence.top and geofence.checkInside(*point):
                    #intrusions.append(geofence)
                    # Add geofence ID to unique intrusion dictionary
                    if acid not in cls.unique_intrusions:
                        cls.unique_intrusions[acid] = dict()
                    # get geo_name
                    geo_name = cls.geo_by_id[geo_ids[i]].name
                    # Get closest point
                    p1,p2 = nearest_points(geofence.polybound, Point(traf.lat[idx], traf.lon[idx]))
                    # Do kwikdist
                    intrusion = geo.kwikdist(p1.x, p1.y, p2.x, p2.y) * aero.nm
                    # Check the previous intrusion severity
                    if geo_ids[i] in cls.unique_intrusions[acid]:
                        if cls.unique_intrusions[acid][geo_ids[i]][1] < intrusion:
                            cls.unique_intrusions[acid][geo_ids[i]] = [geo_name, intrusion, p2.x, p2.y, bs.sim.simt]
                    else:
                        cls.unique_intrusions[acid][geo_ids[i]] = [geo_name, intrusion, p2.x, p2.y,  bs.sim.simt]
                    
        bs.traf.geo_intrusions = cls.unique_intrusions
        return


import numpy as np
import bluesky as bs
from bluesky.tools import aero
from bluesky.tools.aero import kts, ft, fpm
from bluesky.traffic.performance.perfbase import PerfBase
from bluesky.traffic.performance.openap import coeff, thrust
from bluesky.traffic.performance.openap import phase as ph


class OpenAP(PerfBase):
    """
    Open-source Aircraft Performance (OpenAP) Model

    Methods:
        create(): initialize new aircraft with performance parameters
        update(): update performance parameters
    """

    def __init__(self):
        super().__init__()

        self.ac_warning = False  # aircraft mdl to default warning
        self.eng_warning = False  # aircraft engine to default warning

        self.coeff = coeff.Coefficient()

        with self.settrafarrays():
            self.lifttype = np.array([])  # lift type, fixwing [1] or rotor [2]
            self.engnum = np.array([], dtype=int)  # number of engines
            self.engthrmax = np.array([])  # static engine thrust
            self.engbpr = np.array([])  # engine bypass ratio
            self.max_thrust = np.array([])  # thrust ratio at current alt spd
            self.ff_coeff_a = np.array([])  # icao fuel flows coefficient a
            self.ff_coeff_b = np.array([])  # icao fuel flows coefficient b
            self.ff_coeff_c = np.array([])  # icao fuel flows coefficient c
            self.engpower = np.array([])  # engine power, rotor ac
            self.cd0_clean = np.array([])  # Cd0, clean configuration
            self.k_clean = np.array([])  # k, clean configuration
            self.cd0_to = np.array([])  # Cd0, takeoff configuration
            self.k_to = np.array([])  # k, takeoff configuration
            self.cd0_ld = np.array([])  # Cd0, landing configuration
            self.k_ld = np.array([])  # k, landing configuration
            self.delta_cd_gear = np.array([])  # landing gear

            self.vminic = np.array([])
            self.vminer = np.array([])
            self.vminap = np.array([])
            self.vmaxic = np.array([])
            self.vmaxer = np.array([])
            self.vmaxap = np.array([])

            self.vminto = np.array([])
            self.hcross = np.array([])
            self.mmo = np.array([])

    def create(self, n=1):
        # cautious! considering multiple created aircraft with same type
        super().create(n)

        actype = bs.traf.type[-1].upper()

        # Check synonym file if not in open ap actypes
        if (actype not in self.coeff.actypes_rotor) and (
            actype not in self.coeff.dragpolar_fixwing
        ):
            if actype in self.coeff.synodict.keys():
                # warn = f"Warning: {actype} replaced by {self.coeff.synodict[actype]}"
                # print(warn)
                # stack.echo(warn)
                actype = self.coeff.synodict[actype]

        # initialize aircraft / engine performance parameters
        # check fixwing or rotor, default to fixwing
        if actype in self.coeff.actypes_rotor:
            self.lifttype[-n:] = coeff.LIFT_ROTOR
            self.mass[-n:] = 0.5 * (
                self.coeff.acs_rotor[actype]["oew"]
                + self.coeff.acs_rotor[actype]["mtow"]
            )
            self.engnum[-n:] = int(self.coeff.acs_rotor[actype]["n_engines"])
            self.engpower[-n:] = self.coeff.acs_rotor[actype]["engines"][0][1]

        else:
            # convert to known aircraft type
            if actype not in self.coeff.actypes_fixwing:
                # warn = f"Warning: {actype} replaced by B744"
                # print(warn)
                # stack.echo(warn)
                actype = "B744"

            # populate fuel flow model
            es = self.coeff.acs_fixwing[actype]["engines"]
            e = es[list(es.keys())[0]]
            coeff_a, coeff_b, coeff_c = thrust.compute_eng_ff_coeff(
                e["ff_idl"], e["ff_app"], e["ff_co"], e["ff_to"]
            )

            self.lifttype[-n:] = coeff.LIFT_FIXWING

            self.Sref[-n:] = self.coeff.acs_fixwing[actype]["wa"]
            self.mass[-n:] = 0.5 * (
                self.coeff.acs_fixwing[actype]["oew"]
                + self.coeff.acs_fixwing[actype]["mtow"]
            )

            self.engnum[-n:] = int(self.coeff.acs_fixwing[actype]["n_engines"])

            self.ff_coeff_a[-n:] = coeff_a
            self.ff_coeff_b[-n:] = coeff_b
            self.ff_coeff_c[-n:] = coeff_c

            all_ac_engs = list(self.coeff.acs_fixwing[actype]["engines"].keys())
            self.engthrmax[-n:] = self.coeff.acs_fixwing[actype]["engines"][
                all_ac_engs[0]
            ]["thr"]
            self.engbpr[-n:] = self.coeff.acs_fixwing[actype]["engines"][
                all_ac_engs[0]
            ]["bpr"]

        # init type specific coefficients for flight envelops
        if actype in self.coeff.limits_rotor.keys():  # rotorcraft
            self.vmin[-n:] = self.coeff.limits_rotor[actype]["vmin"]
            self.vmax[-n:] = self.coeff.limits_rotor[actype]["vmax"]
            self.vsmin[-n:] = self.coeff.limits_rotor[actype]["vsmin"]
            self.vsmax[-n:] = self.coeff.limits_rotor[actype]["vsmax"]
            self.hmax[-n:] = self.coeff.limits_rotor[actype]["hmax"]

            self.vsmin[-n:] = self.coeff.limits_rotor[actype]["vsmin"]
            self.vsmax[-n:] = self.coeff.limits_rotor[actype]["vsmax"]
            self.hmax[-n:] = self.coeff.limits_rotor[actype]["hmax"]

            self.cd0_clean[-n:] = np.nan
            self.k_clean[-n:] = np.nan
            self.cd0_to[-n:] = np.nan
            self.k_to[-n:] = np.nan
            self.cd0_ld[-n:] = np.nan
            self.k_ld[-n:] = np.nan
            self.delta_cd_gear[-n:] = np.nan

        else:
            if actype not in self.coeff.limits_fixwing.keys():
                actype = "B744"

            self.vminic[-n:] = self.coeff.limits_fixwing[actype]["vminic"]
            self.vminer[-n:] = self.coeff.limits_fixwing[actype]["vminer"]
            self.vminap[-n:] = self.coeff.limits_fixwing[actype]["vminap"]
            self.vmaxic[-n:] = self.coeff.limits_fixwing[actype]["vmaxic"]
            self.vmaxer[-n:] = self.coeff.limits_fixwing[actype]["vmaxer"]
            self.vmaxap[-n:] = self.coeff.limits_fixwing[actype]["vmaxap"]

            self.vsmin[-n:] = self.coeff.limits_fixwing[actype]["vsmin"]
            self.vsmax[-n:] = self.coeff.limits_fixwing[actype]["vsmax"]
            self.hmax[-n:] = self.coeff.limits_fixwing[actype]["hmax"]
            self.axmax[-n:] = self.coeff.limits_fixwing[actype]["axmax"]
            self.vminto[-n:] = self.coeff.limits_fixwing[actype]["vminto"]
            self.hcross[-n:] = self.coeff.limits_fixwing[actype]["crosscl"]
            self.mmo[-n:] = self.coeff.limits_fixwing[actype]["mmo"]

            self.cd0_clean[-n:] = self.coeff.dragpolar_fixwing[actype]["cd0_clean"]
            self.k_clean[-n:] = self.coeff.dragpolar_fixwing[actype]["k_clean"]
            self.cd0_to[-n:] = self.coeff.dragpolar_fixwing[actype]["cd0_to"]
            self.k_to[-n:] = self.coeff.dragpolar_fixwing[actype]["k_to"]
            self.cd0_ld[-n:] = self.coeff.dragpolar_fixwing[actype]["cd0_ld"]
            self.k_ld[-n:] = self.coeff.dragpolar_fixwing[actype]["k_ld"]
            self.delta_cd_gear[-n:] = self.coeff.dragpolar_fixwing[actype][
                "delta_cd_gear"
            ]

        # append update actypes, after removing unknown types
        self.actype[-n:] = [actype] * n

        # Update envelope speed limits
        mask = np.zeros_like(self.actype, dtype=bool)
        mask[-n:] = True
        self.vmin[-n:], self.vmax[-n:] = self._construct_v_limits(mask)

    def update(self, dt):
        """Periodic update function for performance calculations."""
        # update phase, infer from spd, roc, alt
        lenph1 = len(self.phase)
        self.phase = ph.get(
            self.lifttype, bs.traf.tas, bs.traf.vs, bs.traf.alt, unit="SI"
        )

        # update speed limits, based on phase change
        self.vmin, self.vmax = self._construct_v_limits()

        idx_fixwing = np.where(self.lifttype == coeff.LIFT_FIXWING)[0]

        # ----- compute drag -----
        # update drage coefficient based on flight phase
        self.cd0[self.phase == ph.GD] = (
            self.cd0_to[self.phase == ph.GD] + self.delta_cd_gear[self.phase == ph.GD]
        )
        self.cd0[self.phase == ph.IC] = self.cd0_to[self.phase == ph.IC]
        self.cd0[self.phase == ph.AP] = self.cd0_ld[self.phase == ph.AP]
        self.cd0[self.phase == ph.CL] = self.cd0_clean[self.phase == ph.CL]
        self.cd0[self.phase == ph.CR] = self.cd0_clean[self.phase == ph.CR]
        self.cd0[self.phase == ph.DE] = self.cd0_clean[self.phase == ph.DE]
        self.cd0[self.phase == ph.NA] = self.cd0_clean[self.phase == ph.NA]

        self.k[self.phase == ph.GD] = self.k_to[self.phase == ph.GD]
        self.k[self.phase == ph.IC] = self.k_to[self.phase == ph.IC]
        self.k[self.phase == ph.AP] = self.k_ld[self.phase == ph.AP]
        self.k[self.phase == ph.CL] = self.k_clean[self.phase == ph.CL]
        self.k[self.phase == ph.CR] = self.k_clean[self.phase == ph.CR]
        self.k[self.phase == ph.DE] = self.k_clean[self.phase == ph.DE]
        self.k[self.phase == ph.NA] = self.k_clean[self.phase == ph.NA]

        rho = aero.vdensity(bs.traf.alt[idx_fixwing])
        vtas = bs.traf.tas[idx_fixwing]
        rhovs = 0.5 * rho * vtas ** 2 * self.Sref[idx_fixwing]
        cl = self.mass[idx_fixwing] * aero.g0 / rhovs
        self.drag[idx_fixwing] = rhovs * (
            self.cd0[idx_fixwing] + self.k[idx_fixwing] * cl ** 2
        )

        # ----- compute maximum thrust -----
        max_thrustratio_fixwing = thrust.compute_max_thr_ratio(
            self.phase[idx_fixwing],
            self.engbpr[idx_fixwing],
            bs.traf.tas[idx_fixwing],
            bs.traf.alt[idx_fixwing],
            bs.traf.vs[idx_fixwing],
            self.engnum[idx_fixwing] * self.engthrmax[idx_fixwing],
        )
        self.max_thrust[idx_fixwing] = (
            max_thrustratio_fixwing
            * self.engnum[idx_fixwing]
            * self.engthrmax[idx_fixwing]
        )

        # ----- compute net thrust -----
        self.thrust[idx_fixwing] = (
            self.drag[idx_fixwing] + self.mass[idx_fixwing] * bs.traf.ax[idx_fixwing]
        )

        # ----- compute duel flow -----
        thrustratio_fixwing = self.thrust[idx_fixwing] / (
            self.engnum[idx_fixwing] * self.engthrmax[idx_fixwing]
        )
        self.fuelflow[idx_fixwing] = self.engnum[idx_fixwing] * (
            self.ff_coeff_a[idx_fixwing] * thrustratio_fixwing ** 2
            + self.ff_coeff_b[idx_fixwing] * thrustratio_fixwing
            + self.ff_coeff_c[idx_fixwing]
        )

        # ----- update max acceleration ----
        self.axmax = self.calc_axmax()

        # TODO: implement thrust computation for rotor aircraft
        # idx_rotor = np.where(self.lifttype==coeff.LIFT_ROTOR)[0]
        # self.thrust[idx_rotor] = 0

        # update bank angle, due to phase change
        self.bank = np.where((self.phase == ph.GD), 15, self.bank)
        self.bank = np.where(
            (self.phase == ph.IC) | (self.phase == ph.CR) | (self.phase == ph.AP),
            35,
            self.bank,
        )

        # ----- debug statements -----
        # print(bs.traf.id)
        # print(self.phase)
        # print(self.thrust.astype(int))
        # print(np.round(self.fuelflow, 2))
        # print(self.drag.astype(int))
        # print(self.currentlimits())
        # print()

    def limits(self, intent_v_tas, intent_vs, intent_h, ax):
        """apply limits on indent speed, vertical speed, and altitude (called in pilot module)

        Args:
            intent_v_tas (float or 1D-array): intent true airspeed
            intent_vs (float or 1D-array): intent vertical speed
            intent_h (float or 1D-array): intent altitude
            ax (float or 1D-array): acceleration
        Returns:
            floats or 1D-arrays: Allowed TAS, Allowed vetical rate, Allowed altitude
        """
        allow_h = np.where(intent_h > self.hmax, self.hmax, intent_h)

        intent_v_cas = aero.vtas2cas(intent_v_tas, allow_h)
        allow_v_cas = np.where((intent_v_cas < self.vmin), self.vmin, intent_v_cas)
        allow_v_cas = np.where(intent_v_cas > self.vmax, self.vmax, allow_v_cas)
        allow_v_tas = aero.vcas2tas(allow_v_cas, allow_h)
        allow_v_tas = np.where(
            aero.vtas2mach(allow_v_tas, allow_h) > self.mmo,
            aero.vmach2tas(self.mmo, allow_h),
            allow_v_tas,
        )  # maximum cannot exceed MMO

        vs_max_with_acc = (1 - ax / self.axmax) * self.vsmax
        allow_vs = np.where(
            (intent_vs > 0) & (intent_vs > self.vsmax), vs_max_with_acc, intent_vs
        )  # for climb with vs larger than vsmax
        allow_vs = np.where(
            (intent_vs < 0) & (intent_vs < self.vsmin), vs_max_with_acc, allow_vs
        )  # for descent with vs smaller than vsmin (negative)
        allow_vs = np.where(
            (self.phase == ph.GD) & (bs.traf.tas < self.vminto), 0, allow_vs
        )  # takeoff aircraft

        # corect rotercraft speed limits
        ir = np.where(self.lifttype == coeff.LIFT_ROTOR)[0]
        allow_v_tas[ir] = np.where(
            (intent_v_tas[ir] < self.vmin[ir]), self.vmin[ir], intent_v_tas[ir]
        )
        allow_v_tas[ir] = np.where(
            (intent_v_tas[ir] > self.vmax[ir]), self.vmax[ir], allow_v_tas[ir]
        )
        allow_vs[ir] = np.where(
            (intent_vs[ir] < self.vsmin[ir]), self.vsmin[ir], intent_vs[ir]
        )
        allow_vs[ir] = np.where(
            (intent_vs[ir] > self.vsmax[ir]), self.vsmax[ir], allow_vs[ir]
        )

        return allow_v_tas, allow_vs, allow_h

    def currentlimits(self, id=None):
        """Get current kinematic performance envelop.

        Args:
            id (int or 1D-array): Aircraft ID(s). Defualt to None (all aircraft).

        Returns:
            floats or 1D-arrays: Min TAS, Max TAS, Min VS, Max VS

        """
        vtasmin = aero.vcas2tas(self.vmin, bs.traf.alt)

        vtasmax = np.minimum(
            aero.vcas2tas(self.vmax, bs.traf.alt), aero.vmach2tas(self.mmo, bs.traf.alt)
        )

        if id is not None:
            return vtasmin[id], vtasmax[id], self.vsmin[id], self.vsmax[id]
        else:
            return vtasmin, vtasmax, self.vsmin, self.vsmax

    def _construct_v_limits(self, mask=True):
        """Compute speed limist base on aircraft model and flight phases

        Args:
            mask: Indices (boolean) for aircraft to construct speed limits for.
                  When no indices are passed, all aircraft are updated.

        Returns:
            2D-array: vmin, vmax
        """
        n = len(self.actype)
        vmin = np.zeros(n)
        vmax = np.zeros(n)

        ifw = np.where(np.logical_and(self.lifttype == coeff.LIFT_FIXWING, mask))[0]
        vminfw = np.zeros(len(ifw))
        vmaxfw = np.zeros(len(ifw))

        # fixwing
        # obtain flight envelope for speed, roc, and alt, based on flight phase

        # --- minimum speed ---
        vminfw = np.where(self.phase[ifw] == ph.NA, 0, vminfw)
        vminfw = np.where(self.phase[ifw] == ph.IC, self.vminic[ifw], vminfw)
        vminfw = np.where(
            (self.phase[ifw] >= ph.CL) | (self.phase[ifw] <= ph.DE), self.vminer[ifw], vminfw
        )
        vminfw = np.where(self.phase[ifw] == ph.AP, self.vminap[ifw], vminfw)
        vminfw = np.where(self.phase[ifw] == ph.GD, 0, vminfw)

        # --- maximum speed ---
        vmaxfw = np.where(self.phase[ifw] == ph.NA, self.vmaxer[ifw], vmaxfw)
        vmaxfw = np.where(self.phase[ifw] == ph.IC, self.vmaxic[ifw], vmaxfw)
        vmaxfw = np.where(
            (self.phase[ifw] >= ph.CL) | (self.phase[ifw] <= ph.DE), self.vmaxer[ifw], vmaxfw
        )
        vmaxfw = np.where(self.phase[ifw] == ph.AP, self.vmaxap[ifw], vmaxfw)
        vmaxfw = np.where(self.phase[ifw] == ph.GD, self.vmaxic[ifw], vmaxfw)

        # rotor
        ir = np.where(np.logical_and(self.lifttype == coeff.LIFT_ROTOR, mask))[0]
        vminr = self.vmin[ir]
        vmaxr = self.vmax[ir]

        vmin[ifw] = vminfw
        vmax[ifw] = vmaxfw
        vmin[ir] = vminr
        vmax[ir] = vmaxr

        if isinstance(mask, bool):
            return vmin, vmax
        return vmin[mask], vmax[mask]

    def calc_axmax(self):
        # accelerations depending on phase and wing type
        axmax_fixwing_ground = 2
        axmax_rotor = 3.5

        axmax = np.zeros(bs.traf.ntraf)

        # fix-wing, in flight
        axmax = (self.max_thrust - self.drag) / self.mass

        # fix-wing, on ground
        axmax[self.phase == ph.GD] = axmax_fixwing_ground

        # drones
        axmax[self.lifttype == coeff.LIFT_ROTOR] = axmax_rotor

        # global minumum acceleration
        axmax[axmax < 0.5] = 0.5

        return axmax

    def show_performance(self, acid):
        return (
            True,
            f"Flight phase: {ph.readable_phase(self.phase[acid])}\n"
            f"Thrust: {self.thrust[acid] / 1000:d} kN\n"
            f"Drag: {self.drag[acid] / 1000:d} kN\n"
            f"Fuel flow: {self.fuelflow[acid]:.2f} kg/s\n"
            f"Speed envelope: [{self.vmin[acid] / kts:d}, {self.vmax[acid] / kts:d}] kts\n"
            f"Vertical speed envelope: [{self.vsmin[acid] / fpm:d}, {self.vsmax[acid] / fpm:d}] fpm\n"
            f"Ceiling: {self.hmax[acid] / ft:d} ft",
        )



""" Node encapsulates the sim process, and manages process I/O. """
from threading import Thread
import zmq
import msgpack
from bluesky import stack
from bluesky.core.walltime import Timer
from bluesky.network.npcodec import encode_ndarray, decode_ndarray

class IOThread(Thread):
    ''' Separate thread for node I/O. '''
    def run(self):
        ''' Implementation of the I/O loop. '''
        ctx = zmq.Context.instance()
        fe_event = ctx.socket(zmq.DEALER)
        fe_stream = ctx.socket(zmq.PUB)
        be_event = ctx.socket(zmq.PAIR)
        be_stream = ctx.socket(zmq.PAIR)
        fe_event.connect('tcp://localhost:10000')
        fe_stream.connect('tcp://localhost:10001')

        be_event.connect('inproc://event')
        be_stream.connect('inproc://stream')
        poller = zmq.Poller()
        poller.register(fe_event, zmq.POLLIN)
        poller.register(be_event, zmq.POLLIN)
        poller.register(be_stream, zmq.POLLIN)

        while True:
            try:
                poll_socks = dict(poller.poll(None))
            except zmq.ZMQError:
                break  # interrupted

            if poll_socks.get(fe_event) == zmq.POLLIN:
                be_event.send_multipart(fe_event.recv_multipart())
            if poll_socks.get(be_event) == zmq.POLLIN:
                msg = be_event.recv_multipart()
                if msg[0] == b'QUIT':
                    break
                fe_event.send_multipart(msg)
            if poll_socks.get(be_stream) == zmq.POLLIN:
                fe_stream.send_multipart(be_stream.recv_multipart())


class Node:
    def __init__(self):
        self.server_id = b''
        self.node_id = b''
        self.running = True
        ctx = zmq.Context.instance()
        self.event_io = ctx.socket(zmq.PAIR)
        self.stream_out = ctx.socket(zmq.PAIR)
        self.poller = zmq.Poller()
        self.iothread = IOThread()

    def event(self, eventname, eventdata, sender_id):
        ''' Event data handler. Reimplemented in Simulation. '''
        print(f'Received {eventname} data from {sender_id}')

    def step(self):
        ''' Perform one iteration step. Reimplemented in Simulation. '''
        pass

    def start(self):
        ''' Starting of main loop. '''
        # Final Initialization
        # Initialization of sockets.
        self.event_io.bind('inproc://event')
        self.stream_out.bind('inproc://stream')
        self.poller.register(self.event_io, zmq.POLLIN)

        # Start the I/O thread, and receive from it this node's ID
        self.iothread.start()
        self.send_event(b'REGISTER')
        self.node_id = self.event_io.recv_multipart()[-1]
        self.server_id = self.node_id[:5]
        print(f'Node started, id={self.node_id}')

        # run() implements the main loop
        self.run()

        # Send quit event to the worker thread and wait for it to close.
        self.event_io.send(b'QUIT')
        self.iothread.join()

    def quit(self):
        ''' Quit the simulation process. '''
        self.running = False

    def run(self):
        ''' Start the main loop of this node. '''
        while self.running:
            # Get new events from the I/O thread
            while self.poll():
                res = self.event_io.recv_multipart()
                sender_id = res[0]
                name = res[1]
                if name == b'QUIT':
                    self.quit()
                else:
                    data = msgpack.unpackb(res[2], object_hook=decode_ndarray, raw=False)
                    self.event(name, data, sender_id)
            # Perform a simulation step
            self.step()

            # Process timers
            Timer.update_timers()

    def poll(self):
        ''' Poll for incoming data from I/O thread '''
        try:
            poll_socks = dict(self.poller.poll(0))
            return poll_socks.get(self.event_io) == zmq.POLLIN
        except zmq.ZMQError:
            return False

    def addnodes(self, count=1):
        self.send_event(b'ADDNODES', count)

    def send_event(self, name, data=None, target=None):
        # On the sim side, target is obtained from the currently-parsed stack command
        self.event_io.send_multipart([stack.sender() or b'*', name, msgpack.packb(data, default=encode_ndarray, use_bin_type=True)])

    def send_stream(self, name, data):
        self.stream_out.send_multipart([name + self.node_id, msgpack.packb(data, default=encode_ndarray, use_bin_type=True)])



import numpy as np
from bluesky.traffic.performance.openap import coeff

NA = 0  # Unknown phase
GD = 1  # Ground
IC = 2  # Initial climb
CL = 3  # Climb
CR = 4  # Cruise
DE = 5  # Descent
AP = 6  # Approach


def readable_phase(ph):
    phases = {
        0: "Unknown phase",
        1: "Ground",
        2: "Initial climb",
        3: "Climb",
        4: "Cruise",
        5: "Descent",
        6: "Approach",
    }
    return phases[ph]


def get(lifttype, spd, roc, alt, unit="SI"):
    ph = np.zeros(len(spd))

    # phase for fixwings
    ph = np.where(lifttype == coeff.LIFT_FIXWING, get_fixwing(spd, roc, alt, unit), ph)

    # phase for rotors
    ph = np.where(lifttype == coeff.LIFT_ROTOR, get_rotor(spd, roc, alt, unit), ph)
    return ph


def get_fixwing(spd, roc, alt, unit="SI"):
    """Get the phase of flight base on aircraft state data

    Args:
    spd (float or 1D array): aircraft speed(s)
    roc (float or 1D array): aircraft vertical rate(s)
    alt (float or 1D array): aricraft altitude(s)
    unit (String):  unit, default 'SI', option 'EP'

    Returns:
    int: phase indentifier

    """

    if unit not in ["SI", "EP"]:
        raise RuntimeError("wrong unit type")

    if unit == "SI":
        spd = spd / 0.514444
        roc = roc / 0.00508
        alt = alt / 0.3048

    ph = np.zeros(len(spd), dtype=int)

    ph[(alt <= 75)] = GD
    ph[(alt >= 75) & (alt <= 1000) & (roc >= 150)] = IC
    ph[(alt >= 75) & (alt <= 1000) & (roc <= -150)] = AP
    ph[(alt >= 1000) & (roc >= 150)] = CL
    ph[(alt >= 1000) & (roc <= -150)] = DE
    ph[(alt >= 10000) & (roc <= 150) & (roc >= -150)] = CR

    return ph


def get_rotor(spd, roc, alt, unit="SI"):
    ph = np.ones(len(spd)) * NA
    return ph



''' BlueSky tiled map GL object. '''
import numpy as np

import bluesky as bs
from bluesky.ui.qtgl import glhelpers as glh
from bluesky.ui.qtgl.glmap import Map
from bluesky.ui.qtgl.tiledtexture import TiledTexture


bs.settings.set_variable_defaults(tilesource='opentopomap')


class TiledMap(Map):
    @classmethod
    def select(cls):
        ''' Echo tile source usage message to console before selecting TiledMap. '''
        tileinfo = bs.settings.tile_sources.get(bs.settings.tilesource)
        if not tileinfo:
            raise KeyError(f'Tile source {bs.settings.tilesource} not found!')
        msg = tileinfo.get('license', f'Tile source: {bs.settings.tilesource}')
        bs.stack.stack(f'ECHO {msg}')
        return super().select()

    ''' Tiled Texture background for radarview. '''
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.map = glh.VertexArrayObject(glh.gl.GL_TRIANGLE_FAN, shader_type='tiled')
        self.texture = TiledTexture(self.glsurface, bs.settings.tilesource)
        self.offsetzoom_loc = 0

    def create(self):
        mapvertices = np.array(
            [-90.0, 540.0, -90.0, -540.0, 90.0, -540.0, 90.0, 540.0], dtype=np.float32)
        
        self.texture.create()
        self.texture.add_bounding_box(-90, -180, 90, 180)
        self.map.create(vertex=mapvertices, texture=self.texture)
        self.offsetzoom_loc = glh.ShaderSet.get_shader(
            'tiled').uniformLocation('offset_scale')
        # Make sure that we have textures on first draw
        self.texture.panzoom(finished=True)

    def draw(self):
        # Send the (possibly) updated global uniforms to the buffer
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_LATLON)

        # --- DRAW THE MAP AND COASTLINES ---------------------------------------------
        # Map and coastlines: don't wrap around in the shader
        self.shaderset.enable_wrap(False)
        shader = glh.ShaderSet.get_shader('tiled')
        shader.bind()
        shader.setUniformValue(self.offsetzoom_loc, *self.texture.offsetscale)
        self.map.draw()
        # Temp coastlines
        Map._instance.draw(skipmap=True)



""" BlueSky sector occupancy count plugin. """
import numpy as np
# Import the global bluesky objects. Uncomment the ones you need
from bluesky import traf, stack #, settings, navdb, traf, sim, scr, tools
from bluesky.tools import areafilter, datalog

# List of sectors known to this plugin.
sectors    = list()
# List of aircraft counted inside registered sectors in previous update step.
previnside = list()

# Data logger for sector occupancy count logfiles
logger     = None

### Initialization function of your plugin. Do not change the name of this
### function, as it is the way BlueSky recognises this file as a plugin.
def init_plugin():
    # Register a sector count logger
    global logger
    logger = datalog.crelog('OCCUPANCYLOG', None, 'Sector count log')

    # Configuration parameters
    config = {
        'plugin_name':     'SECTORCOUNT',
        'plugin_type':     'sim',
        'update_interval': 3.0,
        'update':          update
        }

    stackfunctions = {
        # The command name for your function
        'SECTORCOUNT': [
            # A short usage string. This will be printed if you type HELP <name> in the BlueSky console
            'SECTORCOUNT LIST OR ADD sectorname or REMOVE sectorname',

            # A list of the argument types your function accepts. For a description of this, see ...
            'txt,[txt]',

            # The name of your function in this plugin
            sectorcount,

            # a longer help text of your function.
            'Add/remove/list sectors for occupancy count']
    }

    # init_plugin() should always return these two dicts.
    return config, stackfunctions


### Periodic update functions that are called by the simulation. You can replace
### this by anything, so long as you communicate this in init_plugin

def update():
    mylog = list()
    for idx, name in enumerate(sectors):
        # Perform inside count, and check entering and leaving aircraft
        inside    = areafilter.checkInside(name, traf.lat, traf.lon, traf.alt)
        ids       = set(np.array(traf.id)[inside])
        previds   = previnside[idx]
        arrived   = str.join(', ', ids - previds)
        left      = str.join(', ', previds - ids)
        n_tot     = len(ids)
        n_arrived = len(arrived)
        n_left    = len(left)

        # Add log string to list
        mylog.append('%s, %d' % (name, n_tot))

        # Print to console
        if n_left > 0:
            stack.echo('%s aircraft that have left: %s' % (name, left))
        if n_arrived > 0:
            stack.echo('%s aircraft that have arrived: %s' % (name, arrived))
        if n_left + n_arrived > 0:
            stack.echo('%s occupancy count: %d' % (name, n_tot))
            previnside[idx] = ids

    # Log data if enabled
    logger.log(str.join(', ', mylog))

def sectorcount(sw, name=''):
    if sw == 'LIST':
        if len(sectors) == 0:
            return True, 'No registered sectors available'
        else:
            return True, 'Registered sectors:', str.join(', ', sectors)
    elif sw == 'ADD':
        # Add new sector to list.
        if name in sectors:
            return True, 'Sector %s already registered.' % name
        elif areafilter.hasArea(name):
            # Add new area to the sector list, and add an initial inside count of traffic
            sectors.append(name)
            inside = areafilter.checkInside(name, traf.lat, traf.lon, traf.alt)
            previnside.append(set(np.array(traf.id)[inside]))
            return True, 'Added %s to sector list.' % name
        else:
            return False, "No area found with name '%s', create it first with one of the shape commands" % name

    else:
        # Remove area from sector list
        if name in sectors:
            idx = sectors.index(name)
            sectors.pop(idx)
            previnside.pop(idx)
            return True, 'Removed %s from sector list.' % name
        else:
            return False, "No sector registered with name '%s'." % name



''' BlueSky synthetic conflict generator plugin. '''
import random
import numpy as np
from bluesky import stack, traf, sim
from bluesky.tools.aero import ft, eas2tas


def init_plugin():
    ''' Plugin initialisation function. '''

    # Configuration parameters
    config = {
        # The name of your plugin
        'plugin_name':     'SYNTHETIC',

        # The type of this plugin. For now, only simulation plugins are possible.
        'plugin_type':     'sim',
        }

    # init_plugin() should always return a configuration dict.
    return config


@stack.commandgroup(aliases=('SYNTHETIC',))
def syn():
    ''' SYNTHETIC: Generate synthetic conflict geometries. '''
    return True, ("This is the synthetic traffic scenario module\n"
                  "Possible subcommands: SIMPLE, SIMPLED, DIFG, SUPER, SPHERE, "
                  "MATRIX, FLOOR, TAKEOVER, WALL, ROW, COLUMN, DISP")


@syn.subcommand
def simple():
    ''' SIMPLE: Generate a simple 2-aircraft conflict. '''
    sim.reset()
    traf.cre(acid="OWNSHIP", actype="GENERIC", aclat=-.5, aclon=0,
             achdg=0, acalt=5000 * ft, acspd=200)
    traf.cre(acid="INTRUDER", actype="GENERIC", aclat=0, aclon=.5,
             achdg=270, acalt=5000 * ft, acspd=200)
    return True


@syn.subcommand
def simpled():
    ''' SIMPLED: Generate a simple 2-aircraft conflict
        with random speed and distance. '''
    sim.reset()
    ds = random.uniform(0.92, 1.08)
    dd = random.uniform(0.92, 1.08)
    traf.cre(acid="OWNSHIP", actype="GENERIC", aclat=-.5 * dd, aclon=0,
             achdg=0, acalt=20000 * ft, acspd=200 * ds)
    traf.cre(acid="INTRUDER", actype="GENERIC", aclat=0, aclon=.5 / dd,
             achdg=270, acalt=20000 * ft, acspd=200 / ds)
    return True


@syn.subcommand(name='SUPER')
def gensuper(numac:int):
    ''' SUPER: Generate a circular super conflict.

        Arguments:
        -numac: The number of aircraft in the conflict.
    '''
    sim.reset()

    distance=0.50 #this is in degrees lat/lon, for now
    alt=20000*ft #ft
    spd=200 #kts
    for i in range(numac):
        angle=2*np.pi/numac*i
        acid = "SUP" + str(i)
        traf.cre(acid=acid, actype="SUPER",
                 aclat=distance * -np.cos(angle),
                 aclon=distance * np.sin(angle),
                 achdg=360.0 - 360.0 / numac * i,
                 acalt=alt, acspd=spd)
    return True


@syn.subcommand
def sphere(numac:int):
    ''' SUPER: Generate a spherical super conflict,
        with 3 layers of superconflicts.
        Arguments:
        -numac: The number of aircraft in each layer.
    '''
    sim.reset()

    distance = 0.5  # this is in degrees lat/lon, for now
    distancenm = distance*111319./1852
    alt = 20000  # ft
    spd = 150  # kts
    vs = 4  # m/s
    timetoimpact = distancenm/spd*3600  # seconds
    altdifference = vs*timetoimpact  # m
    midalt = alt
    lowalt = alt-altdifference
    highalt = alt+altdifference
    hispd = eas2tas(spd, highalt)
    mispd = eas2tas(spd, midalt)
    lospd = eas2tas(spd, lowalt)
    hispd = spd
    mispd = spd
    lospd = spd
    for i in range(numac):
        angle = np.pi*(2./numac*i)
        lat = distance*-np.cos(angle)
        lon = distance*np.sin(angle)
        track = np.degrees(-angle)

        acidl = "SPH"+str(i)+"LOW"
        traf.cre(acid=acidl, actype="SUPER", aclat=lat, aclon=lon,
                 achdg=track, acalt=lowalt*ft, acspd=lospd)
        acidm = "SPH"+str(i)+"MID"
        traf.cre(acid=acidm, actype="SUPER", aclat=lat, aclon=lon,
                 achdg=track, acalt=midalt*ft, acspd=mispd)
        acidh = "SPH"+str(i)+"HIG"
        traf.cre(acid=acidh, actype="SUPER", aclat=lat, aclon=lon,
                 achdg=track, acalt=highalt*ft, acspd=hispd)

        idxl = traf.id.index(acidl)
        idxh = traf.id.index(acidh)

        traf.vs[idxl] = vs
        traf.vs[idxh] = -vs

        traf.selvs[idxl] = vs
        traf.selvs[idxh] = -vs

        traf.selalt[idxl] = highalt
        traf.selalt[idxh] = lowalt
    return True


@syn.subcommand
def funnel(size:int):
    ''' FUNNEL: create a funnel conflict. '''
    sim.reset()
    # traf.asas=CASASfunnel.Dbconf(300., 5.*nm, 1000.*ft)
    mperdeg=111319.
    distance=0.90 #this is in degrees lat/lon, for now
    alt=20000 #meters
    spd=200 #kts
    numac=8 #number of aircraft
    for i in range(numac):
        angle=np.pi/2/numac*i+np.pi/4
        acid="SUP"+str(i)
        traf.cre(acid=acid, actype="SUPER",
                 aclat=distance*-np.cos(angle),
                 aclon=distance*-np.sin(angle),
                 achdg=90, acalt=alt, acspd=spd)

    # the factor 1.01 is so that the funnel doesn't collide with itself
    separation=traf.cd.rpz*1.01 #[m]
    sepdeg=separation/np.sqrt(2.)/mperdeg #[deg]

    for f_row in range(1):
        for f_col in range(15):
            opening=(size+1)/2.*separation/mperdeg
            Coldeg=sepdeg*f_col  #[deg]
            Rowdeg=sepdeg*f_row  #[deg]
            acid1="FUNN"+str(f_row)+"-"+str(f_col)
            acid2="FUNL"+str(f_row)+"-"+str(f_col)
            traf.cre(acid=acid1, actype="FUNNEL",
                     aclat=Coldeg+Rowdeg+opening,
                     aclon=-Coldeg+Rowdeg+0.5,
                     achdg=0, acalt=alt, acspd=0)
            traf.cre(acid=acid2, actype="FUNNEL",
                     aclat=-Coldeg-Rowdeg-opening,
                     aclon=-Coldeg+Rowdeg+0.5,
                     achdg=0, acalt=alt, acspd=0)

@syn.subcommand
def matrix(size:int):
    ''' MATRIX: create a conflict with several aircraft
        flying in a matrix formation.
    '''
    sim.reset()
    mperdeg = 111319.
    hsep = traf.cd.rpz  # [m] horizontal separation minimum
    hseplat = hsep/mperdeg
    matsep = 1.1  # factor of extra space in the matrix
    hseplat = hseplat*matsep
    vel = 200  # m/s
    # degrees latlon flown in 5 minutes
    extradist = (vel*1.1)*5*60/mperdeg
    for i in range(size):
        acidn = "NORTH"+str(i)
        traf.cre(acid=acidn, actype="MATRIX",
                 aclat=hseplat*(size-1.)/2+extradist,
                 aclon=(i-(size-1.)/2)*hseplat,
                 achdg=180, acalt=20000*ft, acspd=vel)
        acids = "SOUTH"+str(i)
        traf.cre(acid=acids, actype="MATRIX",
                 aclat=-hseplat*(size-1.)/2-extradist,
                 aclon=(i-(size-1.)/2)*hseplat,
                 achdg=0, acalt=20000*ft, acspd=vel)
        acide = "EAST"+str(i)
        traf.cre(acid=acide, actype="MATRIX",
                 aclat=(i-(size-1.)/2)*hseplat,
                 aclon=hseplat*(size-1.)/2+extradist,
                 achdg=270, acalt=20000*ft, acspd=vel)
        acidw = "WEST"+str(i)
        traf.cre(acid=acidw, actype="MATRIX",
                 aclat=(i-(size-1.)/2)*hseplat,
                 aclon=-hseplat*(size-1.)/2-extradist,
                 achdg=90, acalt=20000*ft, acspd=vel)


@syn.subcommand
def floor():
    ''' FLOOR: create a conflict with several aircraft flying
        in a floor formation.
    '''
    sim.reset()
    mperdeg = 111319.
    altdif = 3000  # ft
    hsep = traf.cd.rpz  # [m] horizontal separation minimum
    floorsep = 1.1  # factor of extra spacing in the floor
    hseplat = hsep/mperdeg*floorsep
    traf.cre(acid="OWNSHIP", actype="FLOOR",
             aclat=-1, aclon=0,
             achdg=90, acalt=(20000+altdif)*ft, acspd=200)
    idx = traf.id.index("OWNSHIP")
    traf.selvs[idx] = -10
    traf.selalt[idx] = 20000-altdif
    for i in range(20):
        acid = "OTH"+str(i)
        traf.cre(acid=acid, actype="FLOOR",
                 aclat=-1, aclon=(i-10)*hseplat,
                 achdg=90, acalt=20000*ft, acspd=200)
    return True


@syn.subcommand
def takeover(numac:int):
    ''' TAKEOVER: create a conflict with several aircraft overtaking eachother
    '''
    sim.reset()
    mperdeg = 111319.
    vsteps = 50  # [m/s]
    for v in range(vsteps, vsteps*(numac+1), vsteps):  # m/s
        acid = "OT"+str(v)
        distancetofly = v*5*60  # m
        degtofly = distancetofly/mperdeg
        traf.cre(acid=acid, actype="OT", aclat=0, aclon=-degtofly,
                 achdg=90, acalt=20000*ft, acspd=v)
    return True


@syn.subcommand
def wall():
    ''' WALL: create a conflict with several aircraft flying in a wall formation
    '''
    sim.reset()

    mperdeg = 111319.
    distance = 0.6  # in degrees lat/lon, for now
    hsep = traf.cd.rpz  # [m] horizontal separation minimum
    hseplat = hsep/mperdeg
    wallsep = 1.1  # factor of extra space in the wall
    traf.cre(acid="OWNSHIP", actype="WALL",
             aclat=0, aclon=-distance,
             achdg=90, acalt=20000*ft, acspd=200)
    for i in range(20):
        acid = "OTHER"+str(i)
        traf.cre(acid=acid, actype="WALL",
                 aclat=(i-10)*hseplat*wallsep, aclon=distance,
                 achdg=270, acalt=20000*ft, acspd=200)

        return True


@syn.subcommand
def row(numac:int, angle:int, radius:float=1.0, alt:'alt'=1e5,
        spd:'spd'=300.0, actype:'txt'='B747'):
    ''' ROW: create a conflict with several aircraft flying in two rows
        angled towards each other.
    '''
    sim.reset()

    mperdeg = 111319.
    hsep = traf.cd.rpz  # [m] horizontal separation minimum
    hseplat = hsep/mperdeg
    matsep = 1.1  # factor of extra space in the formation
    hseplat = hseplat*matsep

    aclat = radius * np.cos(np.deg2rad(angle))  # [deg]
    aclon = radius * np.sin(np.deg2rad(angle))
    latsep = abs(hseplat*np.cos(np.deg2rad(90-angle)))  # [deg]
    lonsep = abs(hseplat*np.sin(np.deg2rad(90-angle)))

    alternate = 1
    for i in range(numac):  # Create a/c
        aclat = aclat+i*latsep*alternate
        aclon = aclon-i*lonsep*alternate
        traf.cre(acid="ANG"+str(i*2), actype=actype,
                 aclat=aclat, aclon=aclon,
                 achdg=180+angle, acalt=alt*ft, acspd=spd)
        traf.cre(acid="ANG"+str(i*2+1), actype=actype,
                 aclat=aclat, aclon=-aclon,
                 achdg=180-angle, acalt=alt*ft, acspd=spd)
        alternate = alternate * -1

    return True


@syn.subcommand
def col(numac:int, angle:int, radius:float=1.0, alt:'alt'=1e5,
        spd:'spd'=300.0, actype:'txt'='B747'):
    ''' COL: create a conflict with several aircraft flying in two columns
        angled towards each other.
    '''
    sim.reset()

    mperdeg = 111319.
    hsep = traf.cd.rpz  # [m] horizontal separation minimum
    hseplat = hsep/mperdeg
    matsep = 1.1  # factor of extra space in the formation
    hseplat = hseplat*matsep

    aclat = radius * np.cos(np.deg2rad(angle))  # [deg]
    aclon = radius * np.sin(np.deg2rad(angle))
    latsep = abs(hseplat*np.cos(np.deg2rad(angle)))  # [deg]
    lonsep = abs(hseplat*np.sin(np.deg2rad(angle)))

    traf.cre(acid="ANG0", actype=actype,
             aclat=aclat, aclon=aclon,
             achdg=180+angle, acalt=alt*ft, acspd=spd)
    traf.cre(acid="ANG1", actype=actype,
             aclat=aclat, aclon=-aclon,
             achdg=180-angle, acalt=alt*ft, acspd=spd)

    for i in range(1, numac):  # Create a/c
        aclat = aclat+latsep
        aclon = aclon+lonsep
        traf.cre(acid="ANG"+str(i*2), actype=actype,
                 aclat=aclat, aclon=aclon,
                 achdg=180+angle, acalt=alt*ft, acspd=spd)
        traf.cre(acid="ANG"+str(i*2+1), actype=actype,
                 aclat=aclat, aclon=-aclon,
                 achdg=180-angle, acalt=alt*ft, acspd=spd)



""" Autocomplete filenames in the BlueSky console."""
import bluesky as bs

previous_g = ''


# Case insensitive file search
def iglob(pattern):
    def either(c):
        return '[%s%s]' % (c.lower(), c.upper()) if c.isalpha() else c
    return list(bs.resource('scenario').glob(''.join(map(either, pattern))))


def reset():
    global previous_g
    previous_g = ''


def complete(cmd):
    """ Command line IC + filename completion function """
    lcmd = cmd.upper().split()
    newcmd = cmd
    displaytext = ''

    if lcmd[0] in ['IC', 'BATCH', 'CALL', 'PCALL']:
        global previous_g
        g = ''
        if len(lcmd) == 2:
            g += lcmd[1].strip()
        files = iglob(g + '*')

        if len(files) > 0:
            if len(files) == 1:
                newcmd = lcmd[0] + ' ' + files[0].name
            elif g == previous_g:
                displaytext += ' '.join(f.name for f in files)

            else:
                previous_g = g
                idx        = len(g)
                while idx < len(files[0].name) and len(files) == len(iglob(g + files[0].name[idx] + '*')):
                    g += files[0].name[idx].upper()
                    idx += 1

                newcmd = lcmd[0] + ' ' + g

    return newcmd, displaytext



''' BlueSky OpenGL map object. '''
import numpy as np

from bluesky.network.sharedstate import ActData
from bluesky.stack import command
from bluesky.ui import palette
from bluesky.ui.qtgl import glhelpers as glh
from bluesky.ui.loadvisuals import load_coastlines
import bluesky as bs

bs.settings.set_variable_defaults(gfx_path='graphics')
palette.set_default_colours(
    background=(0, 0, 0),
    coastlines=(85, 85, 115))


class Map(glh.RenderObject, layer=-100):
    ''' Radar screen map OpenGL object. '''

    # Per remote node attributes
    show_map: ActData[bool] = ActData(True)
    show_coast: ActData[bool] = ActData(True)

    @command
    def showmap(self, flag:bool=None):
        ''' Show/hide satellite map '''
        # TODO: add to SWRAD
        self.show_map = not self.show_map if flag is None else flag

    @command
    def showcoast(self, flag:bool=None):
        ''' Show/hide satellite map '''
        # TODO: add to SWRAD (flag=GEO)
        self.show_coast = not self.show_coast if flag is None else flag

    def __init__(self, parent=None):
        super().__init__(parent)

        self.map = glh.VertexArrayObject(glh.gl.GL_TRIANGLE_FAN)
        self.coastlines = glh.VertexArrayObject(glh.gl.GL_LINES)
        self.coastindices = []
        self.vcount_coast = 0
        self.wraplon_loc = 0

    def create(self):
        ''' Create GL objects. '''
        # ------- Coastlines -----------------------------
        coastvertices, self.coastindices = load_coastlines()
        self.coastlines.create(vertex=coastvertices, color=palette.coastlines)
        self.vcount_coast = len(coastvertices)

        mapvertices = np.array(
            [-90.0, 540.0, -90.0, -540.0, 90.0, -540.0, 90.0, 540.0], dtype=np.float32)
        texcoords = np.array(
            [1, 3, 1, 0, 0, 0, 0, 3], dtype=np.float32)
        self.wraplon_loc = glh.ShaderSet.get_shader(self.coastlines.shader_type).attribs['lon'].loc

        # Load and bind world texture
        max_texture_size = glh.gl.glGetIntegerv(glh.gl.GL_MAX_TEXTURE_SIZE)
        print('Maximum supported texture size: %d' % max_texture_size)
        for i in [16384, 8192, 4096]:
            fname = bs.resource(bs.settings.gfx_path) / f'world.{i}x{i//2}.dds'
            if max_texture_size >= i and fname.exists():
                print(f'Loading texture {fname}')
                self.map.create(vertex=mapvertices,
                                texcoords=texcoords, texture=fname)
                break

    def draw(self, skipmap=False):
        # Send the (possibly) updated global uniforms to the buffer
        self.shaderset.set_vertex_scale_type(self.shaderset.VERTEX_IS_LATLON)

        # --- DRAW THE MAP AND COASTLINES ---------------------------------------------
        # Map and coastlines: don't wrap around in the shader
        self.shaderset.enable_wrap(False)

        if self.show_map and not skipmap:
            self.map.draw()
        
        shaderset = glh.ShaderSet.selected
        # Skip coastlines if set to disabled
        if not self.show_coast:
            return
        if shaderset.data.wrapdir == 0:
            # Normal case, no wrap around
            self.coastlines.draw(
                first_vertex=0, vertex_count=self.vcount_coast)
        else:
            self.coastlines.bind()
            shader = glh.ShaderProgram.bound_shader
            wrapindex = np.uint32(
                self.coastindices[int(shaderset.data.wraplon) + 180])
            if shaderset.data.wrapdir == 1:
                shader.setAttributeValue(self.wraplon_loc, 360.0)
                self.coastlines.draw(
                    first_vertex=0, vertex_count=wrapindex)
                shader.setAttributeValue(self.wraplon_loc, 0.0)
                self.coastlines.draw(
                    first_vertex=wrapindex, vertex_count=self.vcount_coast - wrapindex)
            else:
                shader.setAttributeValue(self.wraplon_loc, -360.0)
                self.coastlines.draw(
                    first_vertex=wrapindex, vertex_count=self.vcount_coast - wrapindex)
                shader.setAttributeValue(self.wraplon_loc, 0.0)
                self.coastlines.draw(
                    first_vertex=0, vertex_count=wrapindex)
